# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from dataclasses import dataclass
from datetime import datetime
from numbers import Real
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import optimize

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model


@dataclass
class HarmonicRegressionParams(Params):
    period: float
    fourier_order: int

    def __post_init__(self) -> None:
        if self.period <= 0:
            msg = f"The provided period must be greater than zero. Value: {self.period}"
            logging.error(msg)
            raise ValueError(msg)

        if self.fourier_order <= 0:
            msg = "The provided fourier order must be greater than zero. " f"Value: {self.fourier_order}"
            logging.error(msg)
            raise ValueError(msg)


class HarmonicRegressionModel(Model[Optional[np.ndarray]]):
    params: Optional[np.ndarray] = None
    harms: Optional[np.ndarray] = None

    def __init__(self, data: TimeSeriesData, params: HarmonicRegressionParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `is_univariate`.
        if not self.data.is_univariate():
            msg = "The provided time series data is not univariate."
            logging.error(msg)
            raise ValueError(msg)
        self.regression_params = params

    def setup_data(self) -> None:
        pass

    def validate_inputs(self) -> None:
        pass

    def fit(self) -> None:
        """Fits harmonic regression to the time series.
        See fit_harmonics for details.
        """
        self.params, self.harms = self.fit_harmonics(
            self.regression_params.period, self.regression_params.fourier_order
        )

    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, dates: pd.Series, *_args: Optional[Any], **_kwargs: Optional[Any]) -> pd.DataFrame:
        """Predicts with harmonic regression values.
         Call fit before calling this function.
        Parameters
        ----------
        dates: dates to compute the predictions for
        Returns
        -------
        Pandas DataFrame with the dates (time) and the
        forecast values (fcst)
        """
        if self.params is None or self.harms is None:
            msg = "Call fit before predict."
            logging.error(msg)
            raise ValueError(msg)
        harmonics = HarmonicRegressionModel.fourier_series(
            dates, self.regression_params.period, self.regression_params.fourier_order
        )
        result = np.sum(self.params * harmonics, axis=1)
        return pd.DataFrame({"time": dates, "fcst": result.tolist()}, copy=False)

    def plot(
        self,
        ax: Optional[plt.Axes] = None,
        figsize: Optional[Tuple[int, int]] = None,
        title: Optional[str] = "Time Series v.s Fitted Harmonics",
        **kwargs: Optional[Any],
    ) -> plt.Axes:
        """Demeans the time series, fits the harmonics,
            returns the plot and error metrics.
        Parameters
        ----------
        Returns
            Plot of the original time series and the fitted harmonics
            Dataframe with mean square error and absolute error
        """

        if self.params is None or self.harms is None:
            msg = "Call fit before plot."
            logging.error(msg)
            raise ValueError(msg)

        fitted_harmonics = np.sum(self.params * self.harms, axis=1)

        if ax is None:
            if figsize is None:
                figsize = (20, 6)
            _, ax = plt.subplots(figsize=figsize)

        ax.set_title(title)
        ax.set_xlabel("Time")
        ax.set_ylabel("Value")

        # pyre-fixme[16]: `Optional` has no attribute `time`.
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        (line1,) = ax.plot(self.data.time, self.data.value, label="Original signal")
        (line2,) = ax.plot(self.data.time, fitted_harmonics, label="Fitted harmonics")
        ax.legend(handles=[line1, line2])

        return ax

    @staticmethod
    def fourier_series(dates: pd.Series, period: float, series_order: int) -> np.ndarray:
        """Provides Fourier series components with the specified frequency
        and order. The starting time is always the epoch.
        Parameters
        ----------
        dates: pd.Series containing timestamps.
        period: Number of hours of the period.
        series_order: Number of components.
        Returns
        -------
        Matrix with seasonality features.
        """
        # convert to days since epoch
        t = np.array((dates - datetime(1970, 1, 1)).dt.total_seconds().astype(np.float)) / 3600.0
        return np.column_stack(
            [fun((2.0 * (i + 1) * np.pi * t / period)) for i in range(series_order) for fun in (np.sin, np.cos)]
        )

    @staticmethod
    def make_harm_eval(
        harmonics: np.ndarray,
    ) -> Callable[..., np.ndarray]:
        """Defines evaluation function for the optimizer
        Parameters
        ----------
        harmonics: the harmonics to fit
        Returns
        -------
        The evaluation function for the optimizer
        """

        def harm_eval(step: np.ndarray, *params: Real) -> np.ndarray:
            params_ = np.array(params)
            mul = np.multiply(harmonics[step.astype(int)], params_)
            return np.sum(mul, axis=1)

        return harm_eval

    def fit_harmonics(self, period: float, fourier_order: int) -> Tuple[np.ndarray, np.ndarray]:
        """Performs harmonic regression.
        Harmonic regression fits cosines
        amplitude*cos(freq*t + phase). Using double angle identity formulas,
        we have:
        beta1*cos(freq*t) + beta2*sin(freq*t). Thus, we can fit two coefficients,
        which will take care of the amplitude and the phase. If we generate the
        raw cos(freq*t) and sin(freq*t) for each freq we want to have,
        it becomes a linear regression. Since we ignore
        intercept, we demean the time series before fitting.

        Since the regression takes care of the phase, we can pick time 0 wherever
        we want, we just have to use the same for training, test, validation,
        and prediction. We pick that as the epoch; so when we generate
        the raw cos and sin values for the test set,
        and apply the parameters from the training, it will have the right phase.

        Parameters
        ----------
        period: float; seasonality in hours; e.g. 24 for daily
        fourier_order: int; number of harmonics for the given frequency
        harms: externally computed harmonics
        Returns:
            params: coefficients
            harms: feature matrix the generated raw cos and sin;
            for each fourier_order, there is one cos-sin pair.
            Number of colums: fourier_order*2
        """
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        time_series = self.data.value

        harms = HarmonicRegressionModel.fourier_series(
            # pyre-fixme[16]: `Optional` has no attribute `time`.
            self.data.time,
            period,
            fourier_order,
        )

        steps = np.array(list(range(0, len(time_series.index))))
        demeaned = time_series - time_series.mean()

        params, params_covariance = optimize.curve_fit(
            HarmonicRegressionModel.make_harm_eval(harms),
            steps,
            demeaned.tolist(),
            # This is to set the number of params via the number of initial values
            p0=[0.001] * (fourier_order * 2),
        )
        return params, harms



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from datetime import timedelta
from unittest import TestCase

import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import simple_heuristic_model
from kats.models import theta
from kats.models.stlf import STLFModel
from kats.models.stlf import STLFParams
from kats.utils.simulator import Simulator


def load_data_std_cols(path: str) -> pd.DataFrame:
    df = load_data(path)
    df.columns = ["time", "y"]
    return df


METHODS = ["theta", "linear", "quadratic", "simple"]


class testSTLFModel(TestCase):
    def setUp(self) -> None:

        sim = Simulator(n=3 * 144, freq="10T", start=pd.to_datetime("2021-01-01"))  # 3 days of data
        sim.add_trend(magnitude=1)
        sim.add_seasonality(magnitude=50, period=timedelta(days=1))
        sim.add_noise(magnitude=0)
        dense_dates_10min_df = sim.stl_sim().to_dataframe()
        dense_dates_10min_df.rename(columns={"value": "y"}, inplace=True)

        # Load data for each test
        self.TEST_DATA = {
            "daily": {
                "ts": TimeSeriesData(load_data_std_cols("peyton_manning.csv")),
            },
            "multi": {
                "ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
            },
            "dens_ds_10min": {
                "ts": TimeSeriesData(dense_dates_10min_df),
            },
        }

    # pyre-fixme[56]
    @parameterized.expand([("daily", m) for m in METHODS])
    def test_fit_forecast(self, dataset: str, method: str, steps: int = 5) -> None:
        ts = self.TEST_DATA[dataset]["ts"]
        params = STLFParams(m=12, method=method)
        train, truth = ts[:-steps], ts[-steps:]
        m = STLFModel(train, params)
        m.fit()
        pred = m.predict(steps=steps).iloc[:, 1:].to_numpy()

        # check whether the values are close and shapes are correct
        truth = truth.to_dataframe().y.to_numpy()
        self.assertTrue((truth - pred[:, 1]).max() < 2)  # check actual vs true
        self.assertTrue(all(pred[:, 2] >= pred[:, 0]))  # check upper > lower bounds

    # pyre-fixme[56]
    @parameterized.expand([("multi", m) for m in METHODS])
    def test_invalid_predict_length(self, dataset: str, method: str) -> None:
        ts = self.TEST_DATA[dataset]["ts"]
        params = STLFParams(m=10000, method=method)
        self.assertRaises(
            ValueError,
            STLFModel,
            ts,
            params,
        )

    def test_invalid_params(self) -> None:
        self.assertRaises(
            ValueError,
            STLFParams,
            method="random_model",
            m=12,
        )

    # pyre-fixme[56]
    @parameterized.expand([("multi", m) for m in METHODS])
    def test_model_param(self, dataset: str, method: str) -> None:
        ts = self.TEST_DATA[dataset]["ts"]
        params = STLFParams(m=12, method=method)
        params.validate_params()
        self.assertRaises(
            ValueError,
            STLFModel,
            ts,
            params,
        )

    # pyre-fixme[56]
    @parameterized.expand([("daily", m) for m in METHODS])
    def test_str(self, dataset: str, method: str) -> None:
        ts = self.TEST_DATA[dataset]["ts"]
        params = STLFParams(m=12, method=method)
        params.validate_params()
        m = STLFModel(ts, params)
        self.assertEqual(m.__str__(), "STLF")

    # pyre-fixme[56]
    @parameterized.expand([("daily", m) for m in METHODS])
    def test_fit_forecast_no_default_params(self, dataset: str, method: str, steps: int = 5) -> None:
        ts = self.TEST_DATA[dataset]["ts"]
        if method == "theta":
            method_params = theta.ThetaParams(m=2)
        elif method == "linear":
            method_params = linear_model.LinearModelParams(alpha=0.01)
        elif method == "simple":
            method_params = simple_heuristic_model.SimpleHeuristicModelParams()
        else:
            method_params = quadratic_model.QuadraticModelParams(alpha=0.05)
        params = STLFParams(m=12, method=method, method_params=method_params)
        train, truth = ts[:-steps], ts[-steps:]
        m = STLFModel(train, params)
        m.fit()
        pred = m.predict(steps=steps).iloc[:, 1:].to_numpy()

        # check whether the values are close and shapes are correct
        truth = truth.to_dataframe().y.to_numpy()
        self.assertTrue((truth - pred[:, 1]).max() < 2)  # check actual vs true
        self.assertTrue(all(pred[:, 2] >= pred[:, 0]))  # check upper > lower bounds

    # pyre-fixme[56]
    @parameterized.expand(
        [("daily", m, "additive") for m in METHODS] + [("daily", m, "multiplicative") for m in METHODS]
    )
    def test_fit_forecast_decomposition_parameter(
        self, dataset: str, method: str, decomposition_method: str, steps: int = 5
    ) -> None:

        ts = self.TEST_DATA[dataset]["ts"]
        train, truth = ts[:-steps], ts[-steps:]
        params = STLFParams(
            m=12,
            method=method,
            decomposition=decomposition_method,
        )
        m = STLFModel(train, params)
        m.fit()
        pred = m.predict(steps=steps).iloc[:, 1:].to_numpy()
        truth = truth.to_dataframe().y.to_numpy()

        self.assertTrue((truth - pred[:, 1]).max() < 2)  # check actual vs true
        self.assertTrue(all(pred[:, 2] >= pred[:, 0]))  # check upper > lower bounds

    def test_fit_forecast_simple_model(self) -> None:

        steps = 7

        ts = self.TEST_DATA["daily"]["ts"]
        train, truth = ts[:-steps], ts[-steps:]
        params = STLFParams(
            m=12,
            method="simple",
            decomposition="additive",
        )
        m = STLFModel(train, params)
        m.fit()
        pred = m.predict(steps=steps).iloc[:, 1:].to_numpy()
        truth = truth.to_dataframe().y.to_numpy()

        self.assertTrue((truth - pred[:, 1]).max() < 2)  # check actual vs true
        self.assertTrue(all(pred[:, 2] >= pred[:, 0]))  # check upper > lower bounds

    # pyre-fixme[56]
    @parameterized.expand([("dens_ds_10min", m) for m in METHODS])
    def test_fit_forecast_minute(self, dataset: str, method: str, steps: int = 144) -> None:

        ts = self.TEST_DATA[dataset]["ts"]
        params = STLFParams(m=144, method=method, decomposition="additive")
        train, truth = ts[:-steps], ts[-steps:]
        m = STLFModel(train, params)
        m.fit()
        pred = m.predict(steps=steps)

        # check whether the values are close and shapes are correct
        truth = truth.value.values
        self.assertTrue((truth - pred["fcst"]).max() < 4)  # check actual vs true
        self.assertTrue(all(pred["fcst_upper"] >= pred["fcst_lower"]))  # check upper > lower bounds


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""The Holt Winters model is a time series forecast model that applies exponential smoothing three times, it serves as the extension of the simple exponential smoothing forecast model.

More details about the different exponential smoothing model can be found here:
https://en.wikipedia.org/wiki/Exponential_smoothing
In this module we adopt the Holt Winters model implementation from the statsmodels package, full details can be found as follows:
https://www.statsmodels.org/dev/generated/statsmodels.tsa.holtwinters.ExponentialSmoothing.html
We rewrite the corresponding API to accommodate the Kats development style
"""

import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pandas as pd
from statsmodels.tsa.holtwinters import ExponentialSmoothing as HoltWinters
from statsmodels.tsa.holtwinters import HoltWintersResults

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model
from kats.utils.emp_confidence_int import EmpConfidenceInt
from kats.utils.parameter_tuning_utils import (
    get_default_holtwinters_parameter_search_space,
)


class HoltWintersParams(Params):
    """Parameter class for the HoltWinters model

    Not all parameters from the statsmodels API have been implemented here, the full list of the parameter can be found:
    https://www.statsmodels.org/dev/generated/statsmodels.tsa.holtwinters.ExponentialSmoothing.html

    Attributes:
        trend: Optional; A string that specifies the type of trend component. Can be 'add' and 'mul' or 'additive' and 'multiplicative'. Default is 'add'
        damped: Optional; A boolean indicates whether the trend should be damped or not. Default is False
        seasonal: Optional; A string that specifies the type of seasonal component Can be 'add' and 'mul' or 'additive' and 'multiplicative'. Default is None
        seasonal_periods: Optional; An integer that specifies the period for the seasonal component, e.g. 4 for quarterly data and 7 for weekly seasonality of daily data. Default is None
    """

    trend: Optional[str]
    damped: Optional[bool]
    seasonal: Optional[str]
    seasonal_periods: Optional[int]

    __slots__ = ["trend", "damped", "seasonal", "seasonal_periods"]

    def __init__(
        self,
        trend: str = "add",
        damped: bool = False,
        seasonal: Optional[str] = None,
        seasonal_periods: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.trend = trend
        self.damped = damped
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods
        self.validate_params()
        logging.debug(
            "Initialized HoltWintersParams with parameters. "
            "trend:{trend},\
            damped:{damped},\
            seasonal:{seasonal},\
            seasonal_periods{seasonal_periods}".format(
                trend=trend,
                damped=damped,
                seasonal=seasonal,
                seasonal_periods=seasonal_periods,
            )
        )

    def validate_params(self) -> None:
        """Validate the types and values of the input parameters
        Args:
            None

        Returns:
            None
        """

        if self.trend not in ["add", "mul", "additive", "multiplicative", None]:
            msg = "trend parameter is not valid!\
                         use 'add' or 'mul' instead!"
            logging.error(msg)
            raise ValueError(msg)

        if self.seasonal not in ["add", "mul", "additive", "multiplicative", None]:
            msg = "seasonal parameter is not valid!\
                         use 'add' or 'mul' instead!"
            logging.error(msg)
            raise ValueError(msg)


class HoltWintersModel(Model[HoltWintersParams]):
    """Model class for the HoltWinters model

    Attributes:
        data: :class:`kats.consts.TimeSeriesData`, the input historical time series data from TimeSeriesData
        params: The HoltWinters model parameters from HoltWintersParams
    """

    model: Optional[HoltWintersResults] = None
    freq: Optional[str] = None
    dates: Optional[pd.DatetimeIndex] = None
    alpha: Optional[float] = None
    y_fcst: Optional[pd.Series] = None
    fcst_df: Optional[pd.DataFrame] = None

    def __init__(self, data: TimeSeriesData, params: HoltWintersParams) -> None:

        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

    def fit(self, **kwargs: Any) -> None:
        """Fit the model with the specified input parameters"""

        logging.debug("Call fit() with parameters:{kwargs}".format(kwargs=kwargs))
        holtwinters = HoltWinters(
            # pyre-fixme[16]: `Optional` has no attribute `value`.
            self.data.value,
            trend=self.params.trend,
            damped=self.params.damped,
            seasonal=self.params.seasonal,
            seasonal_periods=self.params.seasonal_periods,
        )
        self.model = holtwinters.fit()
        logging.info("Fitted HoltWinters.")

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, include_history: bool = False, **kwargs: Any) -> pd.DataFrame:
        """Predict with fitted HoltWinters model

        If the alpha keyword argument is specified, an empirical confidence interval is computed through a K-fold cross validation and a linear regression model, the forecast outcome will include a confidence interval there; otherwise no confidence interval is included in the final forecast. Please refer to the 'emp_confidence_int' module for full detailed implementation of the empirical confidence interval computation

        Args:
            steps:
            include_history:
        Returns:
            A pd.DataFrame with the forecast and confidence interval (if empirical confidence interval calculation is triggered)
        """
        model = self.model
        if model is None:
            raise ValueError("Call fit() before predict().")

        logging.debug(
            "Call predict() with parameters. " "steps:{steps}, kwargs:{kwargs}".format(steps=steps, kwargs=kwargs)
        )
        if "freq" not in kwargs:
            # pyre-fixme[16]: `Optional` has no attribute `time`.
            self.freq = pd.infer_freq(self.data.time)
        else:
            self.freq = kwargs["freq"]
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]  # Return correct number of periods
        self.include_history = include_history

        if "alpha" in kwargs:
            self.alpha = alpha = kwargs["alpha"]
            # build empirical CI
            error_methods = kwargs.get("error_methods", ["mape"])
            train_percentage = kwargs.get("train_percentage", 70)
            test_percentage = kwargs.get("test_percentage", 10)
            # pyre-fixme[6]: For 1st param expected `Sized` but got
            #  `Optional[TimeSeriesData]`.
            sliding_steps = kwargs.get("sliding_steps", len(self.data) // 5)
            multi = kwargs.get("multi", True)
            eci = EmpConfidenceInt(
                error_methods=error_methods,
                # pyre-fixme[6]: For 2nd param expected `TimeSeriesData` but got
                #  `Optional[TimeSeriesData]`.
                data=self.data,
                params=self.params,
                train_percentage=train_percentage,
                test_percentage=test_percentage,
                sliding_steps=sliding_steps,
                model_class=HoltWintersModel,
                confidence_level=1 - alpha,
                multi=False,
            )
            logging.debug(
                "Use EmpConfidenceInt for CI with parameters: error_methods = "
                f"{error_methods}, train_percentage = {train_percentage}, "
                f"test_percentage = {test_percentage}, sliding_steps = "
                f"{sliding_steps}, confidence_level = {1-alpha}, multi={multi}."
            )
            fcst = eci.get_eci(steps=steps)
            self.y_fcst = fcst["fcst"]
        else:
            fcst = model.forecast(steps)
            self.y_fcst = fcst
            fcst = pd.DataFrame({"time": self.dates, "fcst": fcst}, copy=False)
        logging.info("Generated forecast data from Holt-Winters model.")

        if include_history:
            history_fcst = model.predict(start=0, end=len(self.data.time) - 1)
            self.fcst_df = fcst = pd.concat(
                [
                    pd.DataFrame(
                        {
                            "time": self.data.time,
                            "fcst": history_fcst,
                        },
                        copy=False,
                    ),
                    fcst,
                ],
                copy=False,
            )
        else:
            self.fcst_df = fcst

        return fcst

    def __str__(self) -> str:
        return "HoltWinters"

    @staticmethod
    # pyre-fixme[15]: `get_parameter_search_space` overrides method defined in
    #  `Model` inconsistently.
    def get_parameter_search_space() -> List[Dict[str, Any]]:
        """Get default HoltWinters parameter search space.

        Args:
            None

        Returns:
            A dictionary with the default HoltWinters parameter search space
        """

        return get_default_holtwinters_parameter_search_space()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase

import matplotlib.pyplot as plt
import pandas as pd
import pytest
import seaborn as sns

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.models.bayesian_var import BayesianVAR
from kats.models.bayesian_var import BayesianVARParams


class testBayesianVARModel(TestCase):
    def setUp(self) -> None:
        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)
        self.params = BayesianVARParams()

    def test_univariate_data(self) -> None:
        univ_df = load_data("air_passengers.csv")
        univ_ts = TimeSeriesData(df=univ_df, time_col_name="ds")
        with self.assertRaises(ValueError):
            _ = BayesianVAR(univ_ts, self.params)

    def test_diff_time_name(self) -> None:
        # Create multivariate time series without 'time' as time column name
        df = load_data("air_passengers.csv")
        df["y_2"] = df.y * 2

        # Create correct DataFrame columns that model should produce
        correct_columns = ["ds", "y", "y_2"]

        # Initialize model and test
        ts = TimeSeriesData(df=df, time_col_name="ds")
        m = BayesianVAR(ts, self.params)
        self.assertEqual(list(m.data.to_dataframe().columns), correct_columns)

    def test_predict_error(self) -> None:
        m = BayesianVAR(self.TSData_multi, self.params)
        with self.assertRaises(ValueError):
            # Model needs to be fit before predict
            m.predict(10)

    def test_predict_zero(self) -> None:
        m = BayesianVAR(self.TSData_multi, self.params)
        with self.assertRaises(ValueError):
            m.predict(0)

    def test_plot_error(self) -> None:
        m = BayesianVAR(self.TSData_multi, self.params)
        with self.assertRaises(ValueError):
            # Must fit() and predict() before plot()
            m.plot()

    def test_sigma_u(self) -> None:
        # Also test verbose predict for code coverage.
        params = BayesianVARParams(p=3)
        m = BayesianVAR(self.TSData_multi, params)
        m.fit()
        m.predict(steps=30, include_history=True, verbose=True)
        self.assertEqual(3, m.k_ar)
        # fmt: off
        expected = pd.DataFrame([
        [ 4.43026366e-05,  2.05347685e-05,  1.81942688e-04,
          4.50075234e-05,  1.86048651e-05,  2.05304755e-06,
          3.55016677e-06, -9.27833097e-05,  4.99988031e-07],
        [ 2.05347685e-05,  2.65322543e-05,  3.16169643e-05,
          7.34176226e-06,  2.38786201e-05,  8.43863345e-06,
          6.53970495e-06, -8.30124023e-05, -2.33583196e-08],
        [ 1.81942688e-04,  3.16169643e-05,  1.33754111e-03,
          5.49729738e-05,  3.16310173e-05,  1.27374356e-06,
         -1.59582259e-05, -3.71395599e-04,  1.61197161e-06],
        [ 4.50075234e-05,  7.34176226e-06,  5.49729738e-05,
          4.27842799e-04,  5.83418411e-06, -2.62717772e-06,
          5.35296733e-05, -1.23409176e-05,  1.29914641e-06],
        [ 1.86048651e-05,  2.38786201e-05,  3.16310173e-05,
          5.83418411e-06,  5.89025512e-05,  1.50549994e-05,
          1.92002465e-05, -8.49113392e-05, -5.78487390e-07],
        [ 2.05304755e-06,  8.43863345e-06,  1.27374356e-06,
         -2.62717772e-06,  1.50549994e-05,  8.51511069e-05,
          5.84851727e-05, -5.63406343e-05, -1.88198418e-07],
        [ 3.55016677e-06,  6.53970495e-06, -1.59582259e-05,
          5.35296733e-05,  1.92002465e-05,  5.84851727e-05,
          1.70683186e-04, -5.78501158e-05,  5.46475324e-07],
        [-9.27833097e-05, -8.30124023e-05, -3.71395599e-04,
         -1.23409176e-05, -8.49113392e-05, -5.63406343e-05,
         -5.78501158e-05,  1.20577180e-03,  7.22586139e-07],
        [ 4.99988031e-07, -2.33583196e-08,  1.61197161e-06,
          1.29914641e-06, -5.78487390e-07, -1.88198418e-07,
          5.46475324e-07,  7.22586139e-07,  1.04254200e-07]
        ],
            columns=self.TSData_multi.value.columns,
            index=self.TSData_multi.value.columns,
        )
        # fmt: on
        assert_frame_equal(expected, m.sigma_u)

    def test_bad_params(
        self,
    ) -> None:
        for p, phi0, phi1, phi2, phi3 in [
            (0, 1, 1, 1, 1),
            (1, 0, 1, 1, 1),
            (1, 1, 0, 1, 1),
            (1, 1, 2, 1, 1),
            (1, 1, 1, 0, 1),
            (1, 1, 1, 1, 0),
        ]:
            params = BayesianVARParams()
            params.p = p
            params.phi_0 = phi0
            params.phi_1 = phi1
            params.phi_2 = phi2
            params.phi_3 = phi3
            with self.assertRaises(ValueError):
                params.validate_params()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from unittest import TestCase

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.detectors.multivariate_detector import MultivariateAnomalyDetectorModel


class TestMultivariateAnomalyDetectorModel(TestCase):
    def setUp(self) -> None:
        self.data = load_data("multivariate_anomaly_simulated_data.csv")
        self.outlier_detector = MultivariateAnomalyDetectorModel()

    def test_response_shape_for_multi_series(self) -> None:
        multi_ts = TimeSeriesData(self.data)
        response_multi_ts = self.outlier_detector.fit_predict(data=multi_ts, historical_data=None)

        self.assertEqual(response_multi_ts.scores.time.shape, multi_ts.time.shape)

        self.assertEqual(response_multi_ts.scores.value.shape, multi_ts.value.shape)

        self.assertEqual(
            # pyre-fixme[16]: Optional type has no attribute `value`.
            response_multi_ts.predicted_ts.value.shape,
            multi_ts.value.shape,
        )

    def test_response_shape_with_historical_data(self) -> None:
        single_ts = TimeSeriesData(self.data)
        historical_ts = TimeSeriesData(self.data)
        single_ts.time = single_ts.time + pd.tseries.offsets.DateOffset(days=len(self.data))
        response = self.outlier_detector.fit_predict(single_ts, historical_ts)

        self.assertTrue(np.array_equal(response.scores.time, single_ts.time))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from enum import Enum
from enum import IntEnum
from functools import lru_cache
from typing import Dict

import numpy as np


class MetricType(Enum):
    """Metrics can be scores, errors, or neither."""

    NONE = 0
    """Neither score nor error"""

    SCORE = 1
    """Larger is better (1 is better than 0)"""

    ERROR = 2
    """Closer to zero is better (0 is better than 1 or -1)"""


class Directionality(IntEnum):
    """Metrics can improve in a direction (up or down) or lack clear direction.

    Non-negative error metrics are negative and vice-versa because lower values
    are also closer to zero. However, metrics that can result in negative values
    cannot be simultaneously NEGATIVE and ERROR.
    """

    NONE = 0
    """Neither positive nor negative."""

    POSITIVE = 1
    """Larger is better  (1 is better than 0)"""

    NEGATIVE = -1
    """Smaller is better (-1 better than 0 better than 1)"""


@dataclass
class MetricMetadata:
    """Metadata about a metric method."""

    type: MetricType
    """The type of metric (score, error, neither)."""

    direction: Directionality
    """The directionality of a metric (positive, negative, or neither is better)."""

    lower_bound: float
    """All values returned by this metric are at least the lower bound."""

    upper_bound: float
    """All values returned by this metric are at most the upper bound."""


_NONNEGATIVE_ERROR_METRIC = MetricMetadata(MetricType.ERROR, Directionality.NEGATIVE, 0.0, np.inf)
_ERROR_RATE_METRIC = MetricMetadata(MetricType.ERROR, Directionality.NEGATIVE, 0.0, 1.0)
_UNBOUNDED_ERROR_METRIC = MetricMetadata(MetricType.ERROR, Directionality.NONE, np.NINF, np.inf)
_UNBOUNDED_METRIC = MetricMetadata(MetricType.NONE, Directionality.NONE, np.NINF, np.inf)
_SCORE_MAX_1 = MetricMetadata(MetricType.SCORE, Directionality.POSITIVE, np.NINF, 1.0)
_SCORE_MIN_0 = MetricMetadata(MetricType.SCORE, Directionality.POSITIVE, 0.0, np.inf)


# These metadata values apply only when unweighted. When unconstrained weights
# are used, most metrics become unbounded.
@lru_cache(None)
def _metadata() -> Dict[str, MetricMetadata]:
    metadata = {
        "continuous_rank_probability_score": _ERROR_RATE_METRIC,
        "cross_entropy": _NONNEGATIVE_ERROR_METRIC,
        "frequency_exceed_relative_threshold": _ERROR_RATE_METRIC,
        "linear_error_in_probability_space": _ERROR_RATE_METRIC,
        "max_error": _UNBOUNDED_ERROR_METRIC,
        "mean_absolute_error": _NONNEGATIVE_ERROR_METRIC,
        "mean_absolute_percentage_error": _NONNEGATIVE_ERROR_METRIC,
        "mean_absolute_scaled_error": _NONNEGATIVE_ERROR_METRIC,
        "mean_error": _UNBOUNDED_ERROR_METRIC,
        "mean_interval_score": _SCORE_MIN_0,
        "mean_percentage_error": _UNBOUNDED_ERROR_METRIC,
        "mean_squared_error": _NONNEGATIVE_ERROR_METRIC,
        "mean_squared_log_error": _NONNEGATIVE_ERROR_METRIC,
        "median_absolute_error": _UNBOUNDED_ERROR_METRIC,
        "median_absolute_percentage_error": _NONNEGATIVE_ERROR_METRIC,
        "r2_score": _SCORE_MAX_1,
        "root_mean_squared_error": _NONNEGATIVE_ERROR_METRIC,
        "root_mean_squared_log_error": _NONNEGATIVE_ERROR_METRIC,
        "root_mean_squared_percentage_error": _NONNEGATIVE_ERROR_METRIC,
        "symmetric_bias": _UNBOUNDED_METRIC,
        "symmetric_mean_absolute_percentage_error": _ERROR_RATE_METRIC,
        "tracking_signal": _UNBOUNDED_ERROR_METRIC,
    }
    # aliases
    metadata["bias"] = metadata["mean_error"]
    metadata["crps"] = metadata["continuous_rank_probability_score"]
    metadata["leps"] = metadata["linear_error_in_probability_space"]
    metadata["mae"] = metadata["mean_absolute_error"]
    metadata["mape"] = metadata["mean_absolute_percentage_error"]
    metadata["mase"] = metadata["mean_absolute_scaled_error"]
    metadata["mdae"] = metadata["median_absolute_error"]
    metadata["mdape"] = metadata["median_absolute_percentage_error"]
    metadata["me"] = metadata["mean_error"]
    metadata["median_absolute_deviation"] = metadata["median_absolute_error"]
    metadata["mean_absolute_deviation"] = metadata["mean_absolute_error"]
    metadata["mis"] = metadata["mean_interval_score"]
    metadata["mpe"] = metadata["mean_percentage_error"]
    metadata["mse"] = metadata["mean_squared_error"]
    metadata["r2"] = metadata["r2_score"]
    metadata["rmse"] = metadata["root_mean_squared_error"]
    metadata["rmsle"] = metadata["root_mean_squared_log_error"]
    metadata["rmspe"] = metadata["root_mean_squared_percentage_error"]
    metadata["sbias"] = metadata["symmetric_bias"]
    metadata["smape"] = metadata["symmetric_mean_absolute_percentage_error"]
    return metadata


def inspect(name: str) -> MetricMetadata:
    """Get metadata about a metric, by name.

    Args:
        name: the metric name

    Returns:
        Metadata about that metric.

    Raises:
        ValueError if the metric name is not a Kats metric.
    """
    metadata = _metadata()
    try:
        return metadata[name]
    except KeyError:
        raise ValueError(f"Unknown metric {name}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest.mock as mock
from unittest import TestCase

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.detectors.dtwcpd import DTWCPDChangePoint
from kats.detectors.dtwcpd import DTWCPDDetector
from kats.detectors.dtwcpd import DTWTimeSeriesTooSmallException


class DTWCPDTest(TestCase):
    first_cp_begin = 100
    first_cp_end = 200
    second_cp_begin = 350

    sigma = 0.05  # std. dev
    num_points = 450

    def test_clear_spike_in_zero_data_yields_cp(self) -> None:
        N = 200
        df = pd.DataFrame(
            {
                "time": pd.date_range("2021-01-01", periods=N, freq="D"),
                "ts1": np.zeros(N),
                "ts2": np.zeros(N),
                "ts3": np.zeros(N),
            }
        )
        df.loc[100, "ts3"] = 500
        ts = TimeSeriesData(df)

        dtw_model = DTWCPDDetector(data=ts, sliding_window_size=50, skip_size=10)

        cps = dtw_model.detector()

        expected_result = [
            DTWCPDChangePoint(
                start_time=pd.Timestamp("2021-03-02 00:00:00", freq="D"),
                end_time=pd.Timestamp("2021-04-20 00:00:00", freq="D"),
                confidence=1e9,
                ts_name="ts3",
            )
        ]

        # Single change point at the desired location.
        self.assertEqual(cps, expected_result)

    def test_no_spike_in_zero_data_yields_no_cp(self) -> None:

        N = 200
        df = pd.DataFrame(
            {
                "time": pd.date_range("2021-01-01", periods=N, freq="D"),
                "ts1": np.zeros(N),
                "ts2": np.zeros(N),
                "ts3": np.zeros(N),
            }
        )
        ts = TimeSeriesData(df)

        dtw_model = DTWCPDDetector(data=ts, sliding_window_size=50, skip_size=10)

        cps = dtw_model.detector()

        self.assertEqual(cps, [])

    def test_two_similar_spikes_in_zero_data_yields_cp(self) -> None:
        N = 200
        df = pd.DataFrame(
            {
                "time": pd.date_range("2021-01-01", periods=N, freq="D"),
                "ts1": np.zeros(N),
                "ts2": np.zeros(N),
                "ts3": np.zeros(N),
                "ts4": np.zeros(N),
                "ts5": np.zeros(N),
            }
        )
        df.loc[100, "ts3"] = 500

        df.loc[150, "ts1"] = 10
        df.loc[50, "ts2"] = 20

        ts = TimeSeriesData(df)

        dtw_model = DTWCPDDetector(data=ts, sliding_window_size=50, skip_size=10)

        cps = dtw_model.detector()

        expected_result = [
            DTWCPDChangePoint(
                start_time=pd.Timestamp("2021-03-02 00:00:00", freq="D"),
                end_time=pd.Timestamp("2021-04-20 00:00:00", freq="D"),
                confidence=mock.ANY,
                ts_name="ts3",
            )
        ]

        self.assertEqual(cps, expected_result)

    def test_wrong_sliding_window_wrong_size(self) -> None:
        N = 200
        df = pd.DataFrame(
            {
                "time": pd.date_range("2021-01-01", periods=N, freq="D"),
                "ts1": np.zeros(N),
                "ts2": np.zeros(N),
                "ts3": np.zeros(N),
                "ts4": np.zeros(N),
                "ts5": np.zeros(N),
            }
        )
        df.loc[100, "ts3"] = 500

        df.loc[150, "ts1"] = 10
        df.loc[50, "ts2"] = 20

        ts = TimeSeriesData(df)

        with self.assertRaises(DTWTimeSeriesTooSmallException):
            dtw_model = DTWCPDDetector(data=ts, sliding_window_size=500, skip_size=100)
            dtw_model.detector()

    def test_one_time_series(self) -> None:
        N = 200
        df = pd.DataFrame(
            {
                "time": pd.date_range("2021-01-01", periods=N, freq="D"),
                "ts1": np.zeros(N),
            }
        )
        df.loc[100, "ts1"] = 500
        ts = TimeSeriesData(df)

        with self.assertRaises(DTWTimeSeriesTooSmallException):
            dtw_model = DTWCPDDetector(data=ts, sliding_window_size=50, skip_size=10)
            dtw_model.detector()

    def test_DTWDistance(self) -> None:
        s1 = [1.0, 2, 3, 4, 5, 6]
        s2 = [3.0, 0, 0, 2, 0, 0]
        w = 2
        distance = DTWCPDDetector.DTWDistance(s1, s2, w)
        self.assertAlmostEqual(distance, 8.660254037844387)

    def test_LB_Keogh(self) -> None:
        s1 = [1.0, 2, 3, 4, 5, 6]
        s2 = [3.0, 0, 0, 2, 0, 0]
        w = 2
        lowerBound = DTWCPDDetector.LB_Keogh(s1, s2, w)
        self.assertAlmostEqual(lowerBound, 5.385164807134504)

    # TODO: test it handles time indices

    # TODO: Test Complex traffic pattern repeats (e.g., with a few increasing spikes etc)

    # TODO: Test non-zero data (with random low level noise in traffic)

    # TODO: what about checking in the same time series?



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import feature_extraction  # noqa  # noqa  # noqa
from . import model_io
from . import nowcasting



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List


class KatasMetaDetectorException(Exception):
    pass


class KatsDetectorsUnimplemented(KatasMetaDetectorException):
    pass


class KatsDetectorUnsupportedAlgoName(KatasMetaDetectorException):
    def __init__(self, algo_name: str, valid_algo_names: List[str]) -> None:
        valid_algo_str = ", ".join(valid_algo_names)
        super().__init__(
            f"unsupported detection algorithm name: {algo_name}, please choose one of those: {valid_algo_str}"
        )


class KatsDetectorHPTIllegalHyperParameter(KatasMetaDetectorException):
    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, algorithm_name: str, idx, hyper_params, e: Exception) -> None:
        hp_str = ",".join([f"{k}:{v}" for k, v in hyper_params.items()])
        super().__init__(
            f"detection algorithm {algorithm_name} can not be initialized with {hp_str} (data_y index: {idx}), throw:\n{e}"
        )


class KatsDetectorHPTTrainError(KatasMetaDetectorException):
    pass


class KatsDetectorHPTModelUsedBeforeTraining(KatasMetaDetectorException):
    pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Dict
from typing import List
from unittest import TestCase

import numpy as np
import pandas as pd

from kats.detectors.cusum_model import CUSUMDetectorModel
from kats.detectors.meta_learning.hpt_tuning import MetaDetectHptSelect
from kats.detectors.meta_learning.hpt_tuning import NNParams
from kats.detectors.meta_learning.hpt_tuning import metadata_detect_reader

BASE_MODELS = ["cusum", "static"]
MODELS_PARAMS: Dict[str, List[str]] = {
    "cusum": [
        "scan_window",
        "historical_window",
        "threshold_low",
        "threshold_high",
    ],
    "static": (
        [
            "threshold_low",
            "threshold_high",
        ]
    ),
}


def gen_synthmetadata(n: int) -> pd.DataFrame:
    data = []
    features = np.random.randn(n * 40).reshape(n, -1)
    for i in range(n):
        hpt_res = {}
        for key in BASE_MODELS:
            if key == "cusum":
                params = {}
                params["scan_window"] = np.random.rand()  # num
                params["historical_window"] = 10  # constant example
                params["threshold_low"] = np.random.choice([1, 2, 3])  # cat
                params["threshold_high"] = np.random.choice([4, 5, 6])  # cat
            else:
                params = {}
                params["threshold_low"] = np.random.rand()  # num
                params["threshold_high"] = np.random.choice([4, 5, 6])  # cat

            hpt_res[key] = (params, np.random.rand())

        feature_dict = {str(k): features[i, k] for k in range(features.shape[1])}
        best_model = BASE_MODELS[np.random.randint(0, len(BASE_MODELS))]
        data.append([hpt_res, feature_dict, best_model])
    return pd.DataFrame(data, columns=["hpt_res", "features", "best_model"])


class TestMetaDetectHptSelect(TestCase):
    def setUp(self) -> None:
        self.synth_meatdata = gen_synthmetadata(500)
        self.meta_data_cusum = metadata_detect_reader(
            rawdata=self.synth_meatdata,
            algorithm_name="cusum",
            params_to_scale_down=set(),
        )
        self.meta_data_static = metadata_detect_reader(
            rawdata=self.synth_meatdata,
            algorithm_name="static",
            params_to_scale_down=set(),
        )
        self.nnparams = NNParams(
            scale=False,
            loss_scale=1.0,
            lr=0.001,
            n_epochs=10,
            batch_size=100,
            method="SGD",
            val_size=0.1,
            momentum=0.9,
        )

    def test_metalearn_flow_cusum(self) -> None:
        datax = self.meta_data_cusum["data_x"]
        datay = self.meta_data_cusum["data_y"]

        mdhs = MetaDetectHptSelect(data_x=datax, data_y=datay, detector_model=CUSUMDetectorModel)

        const_params_dict = mdhs.const_params_dict
        self.assertEqual(len(const_params_dict), 1)
        self.assertEqual(list(const_params_dict.keys())[0], "historical_window")
        self.assertEqual(list(const_params_dict.values())[0], 10)

        self.assertEqual(mdhs._data_y.shape, (500, 3))

        mdhs.train(
            num_idx=["scan_window"],
            cat_idx=["threshold_low", "threshold_high"],
            n_hidden_shared=[20],
            n_hidden_cat_combo=[[5], [5]],
            n_hidden_num=[5],
            nnparams=self.nnparams,
        )

        mdhs.plot()

        res = mdhs.get_hpt_from_features(np.random.normal(0, 1, [10, 40]))
        self.assertEqual(res.shape, (10, 4))

    def test_module_errors(self) -> None:
        datax = self.meta_data_cusum["data_x"]
        datay = self.meta_data_cusum["data_y"]

        mdhs = MetaDetectHptSelect(data_x=datax, data_y=datay, detector_model=CUSUMDetectorModel)

        # provide constant params "historical_window"
        with self.assertRaises(ValueError):
            mdhs.train(
                num_idx=["scan_window", "historical_window"],
                cat_idx=["threshold_low", "threshold_high"],
                n_hidden_shared=[20],
                n_hidden_cat_combo=[[5], [5]],
                n_hidden_num=[5],
                nnparams=self.nnparams,
            )

        # haven't been trained
        with self.assertRaises(AssertionError):
            _ = mdhs.get_hpt_from_features(np.random.normal(0, 1, [10, 40]))

        mdhs.train(
            num_idx=["scan_window"],
            cat_idx=["threshold_low", "threshold_high"],
            n_hidden_shared=[20],
            n_hidden_cat_combo=[[5], [5]],
            n_hidden_num=[5],
            nnparams=self.nnparams,
        )

        # unmatched dimension
        with self.assertRaises(RuntimeError):
            _ = mdhs.get_hpt_from_features(np.random.normal(0, 1, [10, 30]))

        # unmatched format
        with self.assertRaises(IndexError):
            _ = mdhs.get_hpt_from_features(np.random.normal(0, 1, 40))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file defines the BackTester classes for Kats.

Kats supports multiple types of backtesters, including:
  - :class:`BackTesterSimple` (basic train & test backtesting).
  - :class:`BackTesterFixedWindow` (discontinuous train & test data).
  - :class:`BackTesterExpandingWindow` (increasing train window size over
    multiple iterations).
  - :class:`BackTesterRollingWindow` (sliding train & test windows over
    multiple iterations).

This module also supports :class:`CrossValidation` with both expanding and
rolling windows.

For more information, check out the Kats tutorial notebook on backtesting!
"""

import logging
import multiprocessing as mp
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from functools import partial
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

try:
    from typing import Protocol
except ImportError:  # pragma: no cover
    from typing_extensions import Protocol  # pragma: no cover

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.consts import _log_error
from kats.metrics.metrics import CoreMetric
from kats.metrics.metrics import core_metric
from kats.utils.datapartition import DataPartitionBase

if TYPE_CHECKING:
    from kats.models.model import Model


DataPartition = Union[
    List[TimeSeriesData],
    Dict[Union[int, str], TimeSeriesData],
    TimeSeriesData,
]


@dataclass
class BacktesterResult:
    # all raw evaluation results
    raw_errors: Optional[List[pd.DataFrame]]
    # error metrics of each split
    fold_errors: List[Optional[Dict[str, List[float]]]]
    # summary error metrics
    summary_errors: Optional[Dict[str, float]]


class Forecaster(Protocol):
    def __call__(self, train: DataPartition, test: DataPartition) -> pd.DataFrame:
        ...  # pragma: no cover

    """
    Function of fitting a forecasting model with `train` and evaluate the fitted model on `test`.
    Return a pd.DataFrame containing the information for computing evaluation metrics.
    """


class Scorer(Protocol):
    def __call__(self, result: pd.DataFrame) -> Dict[str, float]:
        ...  # pragma: no cover

    """Function for calculating evaluation metrics based on `result`.
    """


def _check_max_core(max_core: Optional[int]) -> int:
    """Helper function for validating core number for multi-processing."""

    total_cores = cpu_count()
    if isinstance(max_core, int) and max_core > 0 and max_core < total_cores:
        core_num = max_core
    else:
        core_num = max((total_cores - 1) // 2, 1)
        logging.warning(f"Input `max_core` = {max_core} is invalid, setting `max_core` = {core_num} instead.")
    return core_num


def _get_scorer(scorer: Union[str, List[str], CoreMetric]) -> Optional[Callable[[pd.DataFrame], Dict[str, float]]]:
    """Helper function for validating `scorer`."""

    if isinstance(scorer, str):
        scorer = [scorer]
    if isinstance(scorer, list):
        methods = []
        for error in scorer:
            try:
                methods.append((error, core_metric(error)))
            except Exception as e:
                msg = f"Unsupported error function {error} with error message {e}."
                _log_error(msg)
        # define scorer function
        # pyre-fixme Incompatible return type [7]: Expected `Optional[typing.Callable[[DataFrame], Dict[str, float]]]` but got `Union[Metric, MultiOutputMetric, WeightedMetric]`.
        def calc_error(result: pd.DataFrame) -> Dict[str, float]:
            errors = {}
            for name, func in methods:
                errors[name] = func(result["y"].values, result["fcst"].values)
            return errors

        return calc_error

    elif callable(scorer):
        # pyre-fixme Incompatible return type [7]: Expected `Optional[typing.Callable[[DataFrame], Dict[str, float]]]` but got `Union[Metric, MultiOutputMetric, WeightedMetric]`.
        return scorer

    msg = "Input `scorer` is invalid."
    _log_error(msg)


def kats_units_forecaster(
    params: Params,
    # pyre-fixme
    model_class: Type,
    train: TimeSeriesData,
    test: TimeSeriesData,
) -> Optional[pd.DataFrame]:
    """
    Forecaster function for uni-time series Kats forecaasting model.
    """
    try:
        model = model_class(data=train, params=params)
        model.fit()
        fcst = model.predict(steps=len(test))
        fcst["y"] = test.value.values
        return fcst
    except Exception as e:
        msg = f"Fail to fit model with `model_class` = {model_class} and `params` = {params} with error message: {e}"
        logging.warning(msg)
    return None


class GenericBacktester(ABC):
    """
    This class defines the module for backtesting for generic forecasting model.

    Attributes:
        datapartition: a datapartition object defining the data partition logic of backtesting.
        forecaster: a callable object following protcol `Forecaster`, used for fitting data and generating forecasts.
        scorer: a strategy to evaluate the performance of forecasting model. Can be a or a list of strings representing the error metrics,
                or a callable following `Scorer` protocol.
        summarizer: a string for the method summarizing error metrics. Can be 'average' (i.e., averaged over all splits) or 'weighted' (i.e., weighted averaged by the size of test size). Default is 'average'.
        raw_errors: a boolean for whether returning raw data on the test sets.
        fold_errors: a boolean for whether returning the error metrics on the test sets.
        multi: a boolean for whether using multi-processing.
        max_core: a integer for the number of cores used by multi-processing. Default is None, which sets `max_core = max((all_available_core-1)//2-1, 1)`.
        error_score: a float for the substitude of error score if it is np.nan. Default is np.nan.

    """

    def __init__(
        self,
        datapartition: DataPartitionBase,
        forecaster: Forecaster,
        scorer: Union[str, CoreMetric, List[Union[str, CoreMetric]], Scorer],
        summarizer: str = "average",
        raw_errors: bool = True,
        fold_errors: bool = True,
        multi: bool = True,
        max_core: Optional[int] = None,
        error_score: float = np.nan,
    ) -> None:

        self.datapartition: DataPartitionBase = datapartition
        # pyre-fixme
        self.scorer: Callable[[pd.DataFrame], Dict[str, float]] = _get_scorer(scorer)
        self.forecaster = forecaster

        self.raw_errors: bool = raw_errors
        self.fold_errors: bool = fold_errors
        self.summarizer: str = summarizer
        self.backtest_result: Optional[BacktesterResult] = None

        self.multi: bool = multi
        self.max_core: int = _check_max_core(max_core) if self.multi else 1
        self.error_score: float = error_score

    def run_backtester(self, data: DataPartition) -> None:
        train_test_data = self.datapartition.split(data)
        logging.info("Successfully finish data partitioning!")

        if self.multi:
            pool = Pool(self.max_core)
            all_res = pool.starmap(self._fit_and_scorer, train_test_data)
            pool.close()
            pool.join()
        else:
            all_res = []
            for train, test in train_test_data:
                tmp_all_res = self._fit_and_scorer(train, test)
                all_res.append(tmp_all_res)
        logging.info("Successfully finish evaluating models on all partitions.")
        self.backtest_result = self._summarize(all_res)
        return None

    def _fit_and_scorer(
        self, train: DataPartition, test: DataPartition
    ) -> Tuple[Optional[pd.DataFrame], Optional[Dict[str, float]]]:
        try:
            res = self.forecaster(train=train, test=test)
            score = self.scorer(res)
            return res, score
        except Exception as e:
            logging.info(f"Failed to fit model and get evaluation with error message: {e}.")
            return None, None

    def _get_fold_errors(
        self, raw_fold_errors: List[Optional[Dict[str, float]]]
    ) -> Tuple[List[str], List[Dict[str, float]]]:

        valid_fold_errors = [t for t in raw_fold_errors if t is not None]
        # get error metric names
        if not valid_fold_errors:
            msg = "No valid evaluation available, i.e., every evaluation failed."
            _log_error(msg)

        error_metrics = list(valid_fold_errors[0].keys())
        base_error_score = {em: self.error_score for em in error_metrics}
        fold_errors = []
        for fd in raw_fold_errors:
            if fd is None:
                fold_errors.append(base_error_score)
            else:
                fold_errors.append(fd)
        return error_metrics, fold_errors

    def _summarize(self, results: List[Tuple[pd.DataFrame, Dict[str, float]]]) -> BacktesterResult:

        if not results:
            _log_error("Fail to get evaluation results!")

        raw_errors = [res[0] for res in results]
        raw_fold_errors = [res[1] for res in results]
        # get error metric names
        # pyre-fixme
        error_metrics, fold_errors = self._get_fold_errors(raw_fold_errors)

        # aggregate error metrics from each fold
        summary_errors = {}
        num_fold = len(fold_errors)
        weight = np.ones(num_fold)
        if self.summarizer == "weighted":
            weight = np.array([len(t) for t in raw_errors])
            weight = weight / np.sum(weight) * num_fold
        for em in error_metrics:
            summary_errors[em] = np.nanmean(np.array([t[em] for t in fold_errors]) * weight)

        backtest_res = BacktesterResult(
            raw_errors=raw_errors if self.raw_errors else None,
            # pyre-fixme
            fold_errors=fold_errors if self.fold_errors else None,
            summary_errors=summary_errors,
        )

        return backtest_res

    def get_errors(self) -> Optional[BacktesterResult]:
        if self.backtest_result is not None:
            return self.backtest_result
        msg = "Please execute function `run_backtester()` first."
        _log_error(msg)


class KatsSimpleBacktester(GenericBacktester):
    """
    Backtester module for single time series Kats forecasting models.

    Attributes:
        datapartition: a datapartition object defining the data partition logic of backtesting.
        scorer: a strategy to evaluate the performance of forecasting model. Can be a or a list of strings representing the error metrics,
                or a callable following `Scorer` protocol.
        model_params: a parameter object for the parameters of the target forecasting model.
        model_class: the model class for the target forecasting model.
        summarizer: a string for the method summarizing error metrics. Can be 'average' (i.e., averaged over all splits) or 'weighted' (i.e., weighted averaged by the size of test size). Default is 'average'.
        raw_errors: a boolean for whether returning raw data on the test sets.
        fold_errors: a boolean for whether returning the error metrics on the test sets.
        multi: a boolean for whether using multi-processing.
        max_core: a integer for the number of cores used by multi-processing. Default is None, which sets `max_core = max((all_available_core-1)//2-1, 1)`.
        error_score: a float for the substitude of error score if it is np.nan. Default is np.nan.
    """

    def __init__(
        self,
        datapartition: DataPartitionBase,
        scorer: Union[str, CoreMetric, List[Union[str, CoreMetric]], Scorer],
        model_params: Params,
        # pyre-fixme
        model_class: Type,
        summarizer: str = "average",
        raw_errors: bool = True,
        fold_errors: bool = True,
        multi: bool = True,
        max_core: Optional[int] = None,
        error_score: float = np.nan,
    ) -> None:
        fcster = partial(kats_units_forecaster, params=model_params, model_class=model_class)
        super(KatsSimpleBacktester, self).__init__(
            datapartition=datapartition,
            scorer=scorer,
            # pyre-fixme
            forecaster=fcster,
            summarizer=summarizer,
            raw_errors=raw_errors,
            fold_errors=fold_errors,
            multi=multi,
            max_core=max_core,
            error_score=error_score,
        )


class BackTesterParent(ABC):
    """
    This class defines the parent functions for various backtesting methods.

    Attributes:
        error_methods: List of strings indicating which errors to calculate
          (see `kats.metrics` for exhaustive list).
        data: :class:`kats.consts.TimeSeriesData` object to perform backtest on.
        params: Parameters to train model with.
        model_class: Defines the model type to use for backtesting.
        multi: Boolean flag to use multiprocessing (if set to True).
        offset: Gap between train/test datasets (default 0).
        results: List of tuples `(training_data, testing_data, trained_model,
          forecast_predictions)` storing forecast results.
        errors: Dictionary mapping the error type to value.
        size: An integer for the total number of datapoints.
        freq: A string representing the (inferred) frequency of the
          `pandas.DataFrame`.

    Raises:
      ValueError: The time series is empty or an invalid error type was passed.
    """

    error_methods: List[Tuple[str, CoreMetric]]
    data: TimeSeriesData
    # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
    #  `typing.Type` to avoid runtime subscripting errors.
    model_class: Type
    params: Params
    multi: bool
    offset: int
    results: List[Tuple[np.ndarray, np.ndarray, "Model[Params]", np.ndarray]]
    errors: Dict[str, float]
    size: int
    freq: Optional[str]
    raw_errors: List[np.ndarray]

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        multi: bool,
        offset: int = 0,
        **kwargs: Any,
    ) -> None:
        self.size = size = len(data.time)
        if not size:
            msg = "Passed an empty time series"
            logging.error(msg)
            raise ValueError(msg)

        self.data = data
        self.model_class = model_class
        self.params = params
        self.multi = multi
        self.offset = offset

        self.results = []
        # Handling frequency
        if "freq" in kwargs:
            self.freq = kwargs["freq"]
        else:
            logging.info("Inferring frequency")
            self.freq = pd.infer_freq(self.data.time)

        self.raw_errors = []

        methods = []
        errors = {}
        for error in error_methods:
            try:
                methods.append((error, core_metric(error)))
                errors[error] = 0.0
            except ValueError:
                msg = f"Unsupported error function {error}"
                logging.error(msg)
                raise ValueError(msg)
        self.errors = errors
        self.error_methods = methods

        logging.info("Instantiated BackTester")
        if kwargs:
            logging.info(
                "Additional arguments: {0}".format((", ".join(["{}={!r}".format(k, v) for k, v in kwargs.items()])))
            )
        logging.info("Model type: {0}".format(self.model_class))
        logging.info("Error metrics: {0}".format(error_methods))

        super().__init__()

    def calc_error(self) -> Optional[float]:
        """
        Calculates all errors in `self.error_methods` and stores them in the
        errors dict.

        Returns:
          The error value. None if the error value does not exist.
        """

        logging.info("Calculating Errors")
        if len(self.results) <= 0:
            logging.error("Empty forecast")
            raise ValueError("No results from forecast")

        # Storing total length of predictions for weighting fold errors
        total_fold_length = sum(result[1].size for result in self.results)

        for result in self.results:
            if len(result) != 4:
                logging.error("Invalid result: {0}".format(result))
                raise ValueError("Invalid result")
            training_inputs, truth, _, predictions = result

            # Storing raw errors
            self.raw_errors.append(truth - predictions)

            if training_inputs.size <= 0:
                logging.error("No training data provided ")
                raise ValueError("Not enough training data")

            if predictions.size <= 0:
                logging.error("No predictions provided")
                raise ValueError("Not enough predictions")

            if truth.size <= 0:
                logging.error("No ground truth data provided")
                raise ValueError("Not enough ground truth data")

            if predictions.size != truth.size:
                logging.error("Unequal amount of labels and predictions")
                raise ValueError("Incorrect dimensionality of predictions & labels")

            for name, method in self.error_methods:
                # Weighting the errors by the relative fold length if
                # predictions are of different sizes
                weight = float(len(truth)) / total_fold_length
                self.errors[name] += weight * method(truth, predictions)

    def _create_model(
        self,
        training_data_indices: Tuple[int, int],
        testing_data_indices: Tuple[int, int],
    ) -> Optional[Tuple[np.ndarray, np.ndarray, "Model[Params]", np.ndarray]]:
        """
        Trains model, evaluates it, and stores results in results list.
        """

        training_data_start, training_data_end = training_data_indices
        testing_data_start, testing_data_end = testing_data_indices
        logging.info("Creating TimeSeries train test objects for split")
        logging.info("Train split of {0}, {1}".format(training_data_start, training_data_end))
        logging.info("Test split of {0}, {1}".format(testing_data_start, testing_data_end))

        if (
            training_data_start < 0
            or training_data_start > self.size
            or training_data_end < 0
            or training_data_end > self.size
        ):
            logging.error("Train Split of {0}, {1} was invalid".format(training_data_start, training_data_end))
            raise ValueError("Invalid training data indices in split")

        if (
            testing_data_start < 0
            or testing_data_start > self.size
            or testing_data_end < 0
            or testing_data_end > self.size
        ):
            logging.error("Test Split of {0}, {1} was invalid".format(testing_data_start, testing_data_end))
            raise ValueError("Invalid testing data indices in split")

        training_data = TimeSeriesData(
            pd.DataFrame(
                {
                    "time": self.data.time[training_data_start:training_data_end],
                    "y": self.data.value[training_data_start:training_data_end],
                }
            )
        )

        testing_data = TimeSeriesData(
            pd.DataFrame(
                {
                    "time": self.data.time[testing_data_start:testing_data_end],
                    "y": self.data.value[testing_data_start:testing_data_end],
                }
            )
        )
        if training_data.value.size <= 0:
            logging.error("No training data provided ")
            raise ValueError("Not enough training data")

        if testing_data.value.size <= 0:
            logging.error("No testing data provided ")
            raise ValueError("Not enough testing data")

        logging.info("Training model")
        train_model = self.model_class(data=training_data, params=self.params)
        train_model.fit()

        logging.info("Making forecast prediction")
        fcst = train_model.predict(steps=testing_data.value.size + self.offset, freq=self.freq)
        train_data_only = np.array(training_data.value)
        truth = np.array(testing_data.value)
        predictions = np.array(fcst["fcst"])
        if self.offset:
            predictions = predictions[self.offset :]

        if not self.multi:
            self.results.append((train_data_only, truth, train_model, predictions))
        else:
            return (train_data_only, truth, train_model, predictions)

    def _build_and_train_models(self, splits: Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]) -> None:
        training_splits, testing_splits = splits

        num_splits = len(training_splits)
        if not self.multi:
            for train_split, test_split in zip(training_splits, testing_splits):
                self._create_model(train_split, test_split)
        else:
            pool = mp.Pool(processes=num_splits)
            futures = [
                pool.apply_async(self._create_model, args=(train_split, test_split))
                for train_split, test_split in zip(training_splits, testing_splits)
            ]
            self.results = results = []
            for fut in futures:
                result = fut.get()
                assert result is not None
                results.append(result)
            pool.close()

    def run_backtest(self) -> None:
        """Executes backtest."""

        self._build_and_train_models(self._create_train_test_splits())
        self.calc_error()

    def get_error_value(self, error_name: str) -> float:
        """Gets requested error value.

        Args:
          error_name: A string of the error whose value should be returned.

        Returns:
          A float of the eror value.

        Raises:
          ValueError: The error name is invalid.
        """

        if error_name in self.errors:
            return self.errors[error_name]
        else:
            logging.error("Invalid error name: {0}".format(error_name))
            raise ValueError("Invalid error name")

    @abstractmethod
    def _create_train_test_splits(
        self,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        raise NotImplementedError()


class BackTesterSimple(BackTesterParent):
    """Defines the functions to execute a simple train/test backtest.

    Attributes:
      train_percentage: A float for the percentage of data used for training.
      test_percentage: A float for the percentage of data used for testing.
      error_methods: List of strings indicating which errors to calculate
        (see `kats.metrics` for exhaustive list).
      data: :class:`kats.consts.TimeSeriesData` object to perform backtest on.
      params: Parameters to train model with.
      model_class: Defines the model type to use for backtesting.
      results: List of tuples `(training_data, testing_data, trained_model,
        forecast_predictions)` storing forecast results.
      errors: Dictionary mapping the error type to value.
      size: An integer for the total number of datapoints.
      error_funcs: Dictionary mapping error name to the
        function that calculates it.
      freq: A string representing the (inferred) frequency of the
        `pandas.DataFrame`.
      raw_errors: List storing raw errors (truth - predicted).

    Raises:
      ValueError: Invalid train and/or test params passed. Or the time series
        is empty.

    Sample Usage:
      >>> df = pd.read_csv("kats/data/air_passengers.csv")
      >>> ts = TimeSeriesData(df=df)
      >>> params = ARIMAParams(p=1, d=1, q=1)
      >>> all_errors = ["mape", "smape", "mae", "mase", "mse", "rmse"]
      >>> backtester = BackTesterSimple(
            error_methods=all_errors,
            data=ts,
            params=params,
            train_percentage=75,
            test_percentage=25,
            model_class=ARIMAModel,
          )
      >>> backtester.run_backtest()
      >>> mape = backtester.get_error_value("mape") # Retrieve MAPE error
    """

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        train_percentage: float,
        test_percentage: float,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        **kwargs: Any,
    ) -> None:
        logging.info("Initializing train/test percentages")
        if train_percentage <= 0:
            logging.error("Non positive training percentage")
            raise ValueError("Invalid training percentage")
        elif train_percentage > 100:
            logging.error("Too large training percentage")
            raise ValueError("Invalid training percentage")
        self.train_percentage = train_percentage
        if test_percentage <= 0:
            logging.error("Non positive test percentage")
            raise ValueError("Invalid test percentage")
        elif test_percentage > 100:
            logging.error("Too large test percentage")
            raise ValueError("Invalid test percentage")
        self.test_percentage = test_percentage

        logging.info("Calling parent class constructor")
        super().__init__(error_methods, data, params, model_class, False, **kwargs)

    def _create_train_test_splits(
        self,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        """Creates train/test folds for the backtest."""

        logging.info("Creating train test splits")
        train_size = _get_absolute_size(self.size, self.train_percentage)
        test_size = _get_absolute_size(self.size, self.test_percentage)

        if train_size <= 0 or train_size >= self.size:
            logging.error("Invalid training size: {0}".format(train_size))
            logging.error("Training Percentage: {0}".format(self.train_percentage))
            raise ValueError("Incorrect training size")

        if test_size <= 0 or test_size >= self.size:
            logging.error("Invalid testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            raise ValueError("Incorrect testing size")

        if train_size + test_size > self.size:
            logging.error("Training and Testing sizes too big")
            logging.error("Training size: {0}".format(train_size))
            logging.error("Training Percentage: {0}".format(self.train_percentage))
            logging.error("Testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            raise ValueError("Incorrect training and testing sizes")

        return [(0, train_size)], [(train_size, train_size + test_size)]


class BackTesterRollingOrigin(BackTesterParent):
    """Defines functions to execute an rolling origin backtest.

    A rolling forecast origin backtest conducts a backtest over multiple
    iterations, wherein each iteration, the "forecasting origin"
    (the location separating training and testing datasets) "slides" forward
    by a fixed amount.

    Hereby, the size of the training dataset is usually increased at each
    iteration, while the test dataset "slides" forward to accommodate.
    However, the size of the training dataset can alternatively be held
    constant, in which case at each iteration the start location of the
    training dataset moves forward by the same amount as the "forecast origin".
    Iterations continue until the complete data set is used to either train
    or test in the final interation.

    This procedure is also known in literature as a rolling origin evaluation
    with a continuously increasing in-sample size (train set) and a constant
    out-sample size (test set).
    For more information, check out the Kats tutorial notebooks!

    Attributes:
      start_train_percentage: A float for the initial percentage of data used
        for training. (The train percentage at the end will be 100 -
        test_percentage)
      test_percentage: A float for the percentage of data used for testing.
        (The test set is taken at sliding positions from start_train_percentage
         up to the end of the dataset - only the last fold is at the very end.)
      expanding_steps: An integer for the number of expanding steps (i.e.
        number of folds).
      error_methods: List of strings indicating which errors to calculate
        (see `kats.metrics` for exhaustive list).
      data: :class:`kats.consts.TimeSeriesData` object to perform backtest on.
      params: Parameters to train model with.
      model_class: Defines the model type to use for backtesting.
      constant_train_size: A bool defining if the training data size should be
        constant instead of expanding it at each iteration (default False).
      multi: A boolean flag to toggle multiprocessing (default True).
      results: List of tuples `(training_data, testing_data, trained_model,
        forecast_predictions)` storing forecast results.
      errors: Dictionary mapping the error type to value.
      size: An integer for the total number of datapoints.
      freq: A string representing the (inferred) frequency of the
        `pandas.DataFrame`.
      raw_errors: List storing raw errors (truth - predicted).

    Raises:
      ValueError: One or more of the train, test, or expanding steps params
        were invalid. Or the time series is empty.

    Sample Usage:
      >>> df = pd.read_csv("kats/data/air_passengers.csv")
      >>> ts = TimeSeriesData(df=df)
      >>> params = ARIMAParams(p=1, d=1, q=1)
      >>> all_errors = ["mape", "smape", "mae", "mase", "mse", "rmse"]
      >>> backtester = BackTesterRollingOrigin(
            error_methods=all_errors,
            data=ts,
            params=params,
            start_train_percentage=50,
            test_percentage=25,
            expanding_steps=3,
            model_class=ARIMAModel,
            constant_train_size=False,
          )
      >>> backtester.run_backtest()
      >>> mape = backtester.get_error_value("mape") # Retrieve MAPE error
    """

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        start_train_percentage: float,
        test_percentage: float,
        expanding_steps: int,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        constant_train_size: bool = False,
        multi: bool = True,
        **kwargs: Any,
    ) -> None:
        logging.info("Initializing train/test percentages")
        if start_train_percentage <= 0:
            logging.error("Non positive start training percentage")
            raise ValueError("Invalid start training percentage")
        elif start_train_percentage > 100:
            logging.error("Too large start training percentage")
            raise ValueError("Invalid end training percentage")
        self.start_train_percentage = start_train_percentage
        if test_percentage <= 0:
            logging.error("Non positive test percentage")
            raise ValueError("Invalid test percentage")
        elif test_percentage > 100:
            logging.error("Too large test percentage")
            raise ValueError("Invalid test percentage")
        self.test_percentage = test_percentage
        if start_train_percentage + test_percentage > 100:
            logging.error("Too large combined train and test percentage")
            raise ValueError("Invalid training and testing percentage combination")  # noqa
        elif start_train_percentage + test_percentage == 100:
            if expanding_steps > 1:
                logging.error(
                    "Too large combined train and test percentage for " "%s expanding steps",
                    expanding_steps,
                )
                raise ValueError(
                    "Invalid training and testing percentage combination "
                    f"given for {expanding_steps} expanding steps"
                )
        if expanding_steps < 0:
            logging.error("Non positive expanding steps")
            raise ValueError("Invalid expanding steps")
        self.expanding_steps = expanding_steps
        self.constant_train_size = constant_train_size

        logging.info("Calling parent class constructor")
        super().__init__(error_methods, data, params, model_class, multi, **kwargs)

    def _create_train_test_splits(
        self,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        """Creates train/test folds for the backtest."""

        logging.info("Creating train test splits")
        start_train_size = _get_absolute_size(self.size, self.start_train_percentage)
        test_size = _get_absolute_size(self.size, self.test_percentage)

        if start_train_size <= 0 or start_train_size >= self.size:
            logging.error("Invalid starting training size: {0}".format(start_train_size))
            logging.error("Start Training Percentage: {0}".format(self.start_train_percentage))
            raise ValueError("Incorrect starting training size")

        if test_size <= 0 or test_size >= self.size:
            logging.error("Invalid testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            raise ValueError("Incorrect testing size")

        if start_train_size + test_size > self.size:
            if start_train_size + test_size > self.size:
                logging.error("Training and Testing sizes too big")
                logging.error("Start Training size: {0}".format(start_train_size))
                logging.error("Start Training Percentage: {0}".format(self.start_train_percentage))
            logging.error("Testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            raise ValueError("Incorrect training and testing sizes")
        elif start_train_size + test_size == self.size:
            if self.expanding_steps > 1:
                logging.error("Training and Testing sizes too big " "for multiple steps")
                logging.error("Start Training size: {0}".format(start_train_size))
                logging.error("Start Training Percentage: {0}".format(self.start_train_percentage))
                logging.error("Testing size: {0}".format(test_size))
                logging.error("Testing Percentage: {0}".format(self.test_percentage))
                logging.error("Expanding steps: {}".format(self.expanding_steps))
                raise ValueError("Incorrect training and testing sizes " "for multiple steps")

        # Handling edge case where only 1 fold is needed (same as BackTesterSimple)
        if self.expanding_steps == 1:
            return (
                [(0, start_train_size)],
                [(start_train_size, start_train_size + test_size)],
            )

        train_splits = []
        test_splits = []
        offsets = _return_fold_offsets(0, self.size - start_train_size - test_size, self.expanding_steps)
        for offset in offsets:
            skip_size = 0
            if self.constant_train_size:
                skip_size = offset
            train_splits.append((skip_size, int(start_train_size + offset)))
            test_splits.append(
                (
                    int(start_train_size + offset),
                    int(start_train_size + offset + test_size),
                )
            )
        return train_splits, test_splits


class BackTesterExpandingWindow(BackTesterRollingOrigin):
    """Defines functions to exectute an expanding window backtest.

    This class will be deprecated soon. Please use `BackTesterRollingOrigin`
    with param `constant_train_size = True`.
    """

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        start_train_percentage: float,
        end_train_percentage: float,
        test_percentage: float,
        expanding_steps: int,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        multi: bool = True,
        **kwargs: Any,
    ) -> None:
        logging.info(
            "BackTesterExpandingWindow will be deprecated. Please use the "
            "updated API found in BackTesterRollingOrigin."
        )
        super().__init__(
            error_methods=error_methods,
            data=data,
            params=params,
            start_train_percentage=start_train_percentage,
            test_percentage=test_percentage,
            expanding_steps=expanding_steps,
            model_class=model_class,
            multi=multi,
            constant_train_size=False,
            **kwargs,
        )


class BackTesterRollingWindow(BackTesterRollingOrigin):
    """Defines functions to execute a rolling window backtest.

    This class will be deprecated soon. Please use `BackTesterRollingOrigin`
    with param `constant_train_size = False`.
    """

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        train_percentage: float,
        test_percentage: float,
        sliding_steps: int,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        multi: bool = True,
        **kwargs: Any,
    ) -> None:
        logging.info(
            "BackTesterRollingWindow will be deprecated. Please use the "
            "updated API found in BackTesterRollingOrigin."
        )
        super().__init__(
            error_methods=error_methods,
            data=data,
            params=params,
            start_train_percentage=train_percentage,
            test_percentage=test_percentage,
            expanding_steps=sliding_steps,
            model_class=model_class,
            multi=multi,
            constant_train_size=True,
            **kwargs,
        )


class BackTesterFixedWindow(BackTesterParent):
    """Defines functions to execute a fixed window ahead backtest.

    A fixed window ahead backtest is similar to a standard (i.e. simple)
    backtest, with the caveat that there is a gap between the train and test
    data sets. The purpose of this type backtest is to focus on the long range
    forecasting ability of the model.

    Attributes:
      train_percentage: A float for the percentage of data used for training.
      test_percentage: A float for the percentage of data used for testing.
      window_percentage: A float for the percentage of data used for the
        fixed window.
      error_methods: List of strings indicating which errors to calculate
        (see `kats.metrics` for exhaustive list).
      data: :class:`kats.consts.TimeSeriesData` object to perform backtest on.
      params: Parameters to train model with.
      model_class: Defines the model type to use for backtesting.
      results: List of tuples `(training_data, testing_data, trained_model,
        forecast_predictions)` storing forecast results.
      errors: Dictionary mapping the error type to value.
      size: An integer for the total number of datapoints.
      freq: A string representing the (inferred) frequency of the
        `pandas.DataFrame`.
      raw_errors: List storing raw errors (truth - predicted).

    Raises:
      ValueError: One or more of the train, test, or fixed window params were
        invalid. Or the time series is empty.

    Sample Usage:
      >>> df = pd.read_csv("kats/data/air_passengers.csv")
      >>> ts = TimeSeriesData(df=df)
      >>> params = ARIMAParams(p=1, d=1, q=1)
      >>> all_errors = ["mape", "smape", "mae", "mase", "mse", "rmse"]
      >>> backtester = BackTesterFixedWindow(
            error_methods=all_errors,
            data=ts,
            params=params,
            train_percentage=50,
            test_percentage=25,
            window_percentage=25,
            model_class=ARIMAModel,
          )
      >>> backtester.run_backtest()
      >>> mape = backtester.get_error_value("mape") # Retrieve MAPE error
    """

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        train_percentage: float,
        test_percentage: float,
        window_percentage: int,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        **kwargs: Any,
    ) -> None:
        logging.info("Initializing train/test percentages")
        if train_percentage <= 0:
            logging.error("Non positive training percentage")
            raise ValueError("Invalid training percentage")
        elif train_percentage > 100:
            logging.error("Too large training percentage")
            raise ValueError("Invalid training percentage")
        self.train_percentage = train_percentage
        if test_percentage <= 0:
            logging.error("Non positive test percentage")
            raise ValueError("Invalid test percentage")
        elif test_percentage > 100:
            logging.error("Too large test percentage")
            raise ValueError("Invalid test percentage")
        self.test_percentage = test_percentage
        if window_percentage < 0:
            logging.error("Non positive window percentage")
            raise ValueError("Invalid window percentage")
        elif window_percentage > 100:
            logging.error("Too large window percentage")
            raise ValueError("Invalid window percentage")
        self.window_percentage = window_percentage

        offset = _get_absolute_size(len(data.time), self.window_percentage)

        logging.info("Calling parent class constructor")
        super().__init__(error_methods, data, params, model_class, False, offset, **kwargs)

    def _create_train_test_splits(
        self,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]]]:
        """Creates train/test folds for the backtest."""

        logging.info("Creating train test splits")
        train_size = _get_absolute_size(self.size, self.train_percentage)
        test_size = _get_absolute_size(self.size, self.test_percentage)
        window_size = _get_absolute_size(self.size, self.window_percentage)

        if train_size <= 0 or train_size >= self.size:
            logging.error("Invalid training size: {0}".format(train_size))
            logging.error("Training Percentage: {0}".format(self.train_percentage))
            raise ValueError("Incorrect training size")

        if test_size <= 0 or test_size >= self.size:
            logging.error("Invalid testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            raise ValueError("Incorrect testing size")

        if train_size + test_size + window_size > self.size:
            logging.error("Combo of Training, Testing, & Window sizes too big")
            logging.error("Training size: {0}".format(train_size))
            logging.error("Training Percentage: {0}".format(self.train_percentage))
            logging.error("Testing size: {0}".format(test_size))
            logging.error("Testing Percentage: {0}".format(self.test_percentage))
            logging.error("Window size: {0}".format(window_size))
            logging.error("Window Percentage: {0}".format(self.window_percentage))
            raise ValueError("Incorrect training, testing, & window sizes")

        train_splits = [(0, int(train_size))]
        test_splits = [(int(train_size + window_size), int(train_size + window_size + test_size))]
        return train_splits, test_splits


class CrossValidation:
    """Defines class to execute time series cross validation.

    Cross validation is a useful technique to use multiple folds of the
    training and testing data to help optimize the performance of the
    model (e.g. hyperparameter tuning). For more info on cross validation, see
    https://en.wikipedia.org/wiki/Cross-validation_(statistics)

    This procedure is also known in literature as a rolling origin evaluation.

    Attributes:
      train_percentage: A float for the percentage of data used for training.
      test_percentage: A float for the percentage of data used for testing.
      num_folds: An integer for the number of folds to use.
      error_methods: List of strings indicating which errors to calculate
        (see `kats.metrics` for exhaustive list).
      data: :class:`kats.consts.TimeSeriesData` object to perform backtest on.
      params: Parameters to train model with.
      model_class: Defines the model type to use for backtesting.
      constant_train_size: A boolean flag to keep the train set size constant,
          sliding it forward with each iteration instead of expanding the
          train set with each iteration (default False).
      multi: A boolean flag to toggle multiprocessing (default True).
      results: List of tuples `(training_data, testing_data, trained_model,
        forecast_predictions)` storing forecast results.
      errors: Dictionary mapping the error type to value.
      size: An integer for the total number of datapoints.
      raw_errors: List storing raw errors (truth - predicted).

    Raises:
      ValueError: One or more of the train, test, or num_folds params
        were invalid. Or the time series is empty.

    Sample Usage:
      >>> df = pd.read_csv("kats/data/air_passengers.csv")
      >>> ts = TimeSeriesData(df=df)
      >>> params = ARIMAParams(p=1, d=1, q=1)
      >>> all_errors = ["mape", "smape", "mae", "mase", "mse", "rmse"]
      >>> cv = CrossValidation(
            error_methods=all_errors,
            data=ts,
            params=params,
            train_percentage=50,
            test_percentage=25,
            num_folds=3,
            model_class=ARIMAModel,
            rolling_window=True
          )
      >>> cv.run_cv()
      >>> mape = cv.get_error_value("mape") # Retrieve MAPE error
    """

    size: int
    results: List[Tuple[np.ndarray, np.ndarray, "Model[Params]", np.ndarray]]
    num_folds: int
    errors: Dict[str, float]
    raw_errors: List[np.ndarray]
    _backtester: BackTesterParent

    def __init__(
        self,
        error_methods: List[str],
        data: TimeSeriesData,
        params: Params,
        train_percentage: float,
        test_percentage: float,
        num_folds: int,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        model_class: Type,
        constant_train_size: bool = False,
        multi: bool = True,
    ) -> None:
        logging.info("Initializing and validating parameter values")
        if train_percentage <= 0:
            logging.error("Non positive training percentage")
            raise ValueError("Invalid training percentage")
        elif train_percentage > 100:
            logging.error("Too large training percentage")
            raise ValueError("Invalid training percentage")
        self.train_percentage = train_percentage
        if test_percentage <= 0:
            logging.error("Non positive test percentage")
            raise ValueError("Invalid test percentage")
        elif test_percentage > 100:
            logging.error("Too large test percentage")
            raise ValueError("Invalid test percentage")
        self.test_percentage = test_percentage
        if num_folds < 0:
            logging.error("Non positive number of folds")
            raise ValueError("Invalid number of folds")
        self.num_folds = num_folds

        self.size = len(data.time)
        if self.size <= 0:
            logging.error("self.size <= 0")
            logging.error("self.size: {0}".format(self.size))
            raise ValueError("Passing an empty time series")

        self.results = []
        self.errors = {}
        self.raw_errors = []

        if not constant_train_size:
            self._backtester = BackTesterExpandingWindow(
                error_methods,
                data,
                params,
                self.train_percentage,
                100 - self.test_percentage,
                self.test_percentage,
                self.num_folds,
                model_class,
                multi=multi,
            )
        else:
            self._backtester = BackTesterRollingWindow(
                error_methods,
                data,
                params,
                self.train_percentage,
                self.test_percentage,
                self.num_folds,
                model_class,
                multi=multi,
            )

    # Run cross validation
    def run_cv(self) -> None:
        """Runs the cross validation."""

        logging.info("Running training and evaluation")
        self._backtester.run_backtest()
        self.results = self._backtester.results
        self.errors = self._backtester.errors
        self.raw_errors = self._backtester.raw_errors
        logging.info("Finished")

    def get_error_value(self, error_name: str) -> float:
        """Gets requested error value.

        Args:
          error_name: A string of the error whose value should be returned.

        Returns:
          A float of the eror value.

        Raises:
          ValueError: The error name is invalid.
        """

        if error_name in self.errors:
            return self.errors[error_name]
        else:
            logging.error("Invalid error name: {0}".format(error_name))
            raise ValueError("Invalid error name")


def _get_absolute_size(size: int, percent: float) -> int:
    """
    Returns absolute size corresponding to percentage of array of length size.
    """

    val = np.floor(size * percent / 100)
    return int(val)


def _return_fold_offsets(start: int, end: int, num_folds: int) -> List[int]:
    """
    Returns approximately even length fold offsets for a given range.
    """

    offsets = [0]
    splits = np.array_split(range(end - start), num_folds - 1)
    for split in splits:
        prev_offset = offsets[-1]
        offsets.append(split.size + prev_offset)
    return offsets



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This module contains some of the key data structures in the Kats library,
including :class:`TimeSeriesData`, :class:`TimeSeriesChangePoint`, and
:class:`TimeSeriesIterator`.

:class:`TimeSeriesChangePoint` is the return type of many of the Kats detection
algorithms.

:class:`TimeSeriesData` is the fundamental data structure in the Kats library,
that gives uses access to a host of forecasting, detection, and utility
algorithms right at the user's fingertips.
"""

from __future__ import annotations

import builtins
import copy
import datetime
import logging
from collections.abc import Iterable
from enum import Enum
from enum import auto
from enum import unique
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import dateutil
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_numeric_dtype
from pandas.tseries.frequencies import to_offset

from kats.compat.pandas import assert_frame_equal
from kats.compat.pandas import assert_series_equal

FigSize = Tuple[int, int]


# Constants
DEFAULT_TIME_NAME = "time"  # Default name for the time column in TimeSeriesData
DEFAULT_VALUE_NAME = "value"  # Default name for the value column in TimeSeriesData
PREFIX_OP_1 = "_kats.1"  # Internal prefix used when merging two TimeSeriesData objects
PREFIX_OP_2 = "_kats.2"  # Second internal prefix used when merging two TimeSeriesData objects
INTERPOLATION_METHODS = {
    "linear",
    "bfill",
    "ffill",
}  # List of possible interpolation methods

IRREGULAR_GRANULARITY_ERROR = (
    "This algorithm or this parameter setup does not support input data with irregular data granularity. "
    "Please update your query to ensure that your data have fixed granularity."
)


def _log_error(msg: str) -> ValueError:
    logging.error(msg)
    return ValueError(msg)


class TimeSeriesChangePoint:
    """Object returned by detector classes.

    Attributes:

        start_time: Start time of the change.
        end_time: End time of the change.
        confidence: The confidence of the change point.
    """

    def __init__(
        self,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        confidence: float,
    ) -> None:
        self._start_time = start_time
        self._end_time = end_time
        self._confidence = confidence

    @property
    def start_time(self) -> pd.Timestamp:
        return self._start_time

    @property
    def end_time(self) -> pd.Timestamp:
        return self._end_time

    @property
    def confidence(self) -> float:
        return self._confidence

    def __repr__(self) -> str:
        return (
            f"TimeSeriesChangePoint(start_time: {self._start_time}, end_time: "
            f"{self._end_time}, confidence: {self._confidence})"
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, TimeSeriesChangePoint)
            and (self._start_time == other._start_time)
            and (self._end_time == other._end_time)
            and (self._confidence == other._confidence)
        )

    def __hash__(self) -> int:
        # Allow subclasses to override __repr__ without affecting __hash__.
        return hash("{self._start_time},{self._end_time},{self._confidence}")


class TimeSeriesData:
    """The fundamental Kats data structure to store a time series.

    In order to access much of the functionality in the Kats library, users
    must initialize the :class:`TimeSeriesData` class with their data first.

    Initialization. :class:`TimeSeriesData` can be initialized from the
    following data sources:

        - `pandas.DataFrame`
        - `pandas.Series`
        - `pandas.DatetimeIndex`

    Typical usage example for initialization:

    >>> import pandas as pd
    >>> df = pd.read_csv("/kats/data/air_passengers.csv")
    >>> ts = TimeSeriesData(df=df, time_col_name="ds")

    Initialization arguments (all optional, but must choose one way to
    initialize e.g. `pandas.DataFrame`):

    - df: A `pandas.DataFrame` storing the time series (default None).
    - sort_by_time: A boolean indicating whether the :class:`TimeSeriesData`
        should be sorted by time (default True).
    - time: a `pandas.Series` or `pandas.DatetimeIndex` storing the time
        values (default None).
    - value: A pandas.Series or pandas.DataFrame storing the series value(s)
        (default None).
    - time_col_name: A string representing the value of the time column (
        default "time")
    - date_format: A string specifying the format of the date/time in the
        time column. Useful for faster parsing, and required
        `pandas.to_datetime()` cannot parse the column otherwise (default None).
    - use_unix_time: A boolean indicating if the time is represented as
        unix time (default False).
    - unix_time_units: A string indicating the units of the unix time -- only
        used if `use_unix_time=True` (default "ns").
    - tz: A string representing the timezone of the time values (default None).
    - tz_ambiguous: A string representing how to handle ambiguous timezones
        (default "raise").
    - tz_nonexistant: A string representing how to handle nonexistant timezone
        values (default "raise").

    Raises:
      ValueError: Invalid params passed when trying to create the
        :class:`TimeSeriesData`.

    Operations. Many operations that you can do with `pandas.DataFrame` objects
    are also applicable to :class:`TimeSeriesData`. For example:

      >>> ts[0:2] # Slicing
      >>> ts_1 == ts_2 # Equality
      >>> ts_1.extend(ts_2) # Extend
      >>> ts.plot(cols=["y"]) # Visualize

    Utility Functions. Many utility functions for converting
    :class:`TimeSeriesData` objects to other common data structures exist.
    For example:

      >>> ts.to_dataframe() # Convert to pandas.DataFrame
      >>> ts.to_array() # Convert to numpy.ndarray

    Attributes:
      time: A `pandas.Series` object storing the time values of the time
        series.
      value: A `pandas.Series` (if univariate) or `pandas.DataFrame` (if
        multivariate) object storing the values of each field in the time
        series.
      min: A float or `pandas.Series` representing the min value(s) of the
        time series.
      max: A float or `pandas.Series` representing the max value(s) of the
        time series.
    """

    _time: pd.Series
    _value: Union[pd.Series, pd.DataFrame]
    _min: float = np.nan
    _max: float = np.nan

    def __init__(  # noqa C901
        self,
        df: Optional[pd.DataFrame] = None,
        sort_by_time: bool = True,
        time: Union[pd.Series, pd.DatetimeIndex, None] = None,
        value: Union[pd.Series, pd.DataFrame, None] = None,
        time_col_name: str = DEFAULT_TIME_NAME,
        date_format: Optional[str] = None,
        use_unix_time: bool = False,
        unix_time_units: str = "ns",
        tz: Optional[str] = None,
        tz_ambiguous: Union[str, np.ndarray] = "raise",
        tz_nonexistent: str = "raise",
    ) -> None:
        """Initializes :class:`TimeSeriesData` class with arguments provided."""
        self.time_col_name = time_col_name

        # If DataFrame is passed
        if df is not None:
            if not isinstance(df, pd.DataFrame):
                msg = "Argument df needs to be a pandas.DataFrame but is of type " f"{type(df)}."
                raise _log_error(msg)
            # If empty DataFrame is passed then create an empty object
            if df.empty:
                self._time = pd.Series([], name=time_col_name, dtype=float)
                self._value = pd.Series([], name=DEFAULT_VALUE_NAME, dtype=float)
                logging.warning("Initializing empty TimeSeriesData object")
            # Otherwise initialize TimeSeriesData from DataFrame
            else:
                # Ensuring time column is present in DataFrame
                if self.time_col_name not in df.columns:
                    msg = f"Time column {self.time_col_name} not in DataFrame"
                    raise _log_error(msg)
                # Parsing time column into correct format
                df = df.copy()
                df.reset_index(inplace=True, drop=True)
                df[self.time_col_name] = self._set_time_format(
                    series=df[self.time_col_name],
                    date_format=date_format,
                    use_unix_time=use_unix_time,
                    unix_time_units=unix_time_units,
                    tz=tz,
                    tz_ambiguous=tz_ambiguous,
                    tz_nonexistent=tz_nonexistent,
                )
                # Sort by time
                df = self._sort_by_time(sort_by_time=sort_by_time, df=df)
                self._extract_from_df(df=df)

        # If separate objects are passed
        elif time is not None and value is not None:
            if not (
                (isinstance(time, pd.core.series.Series) or isinstance(time, pd.DatetimeIndex))
                and (isinstance(value, pd.core.series.Series) or isinstance(value, pd.DataFrame))
            ):
                msg = (
                    f"Invalid types: time is {type(time)} when it must be a "
                    + "pandas.Series or pandas.DatetimeIndex and value is "
                    + f"{type(value)} when it must be a pandas.DataFrame or "
                    + "pandas.Series"
                )
                raise _log_error(msg)
            if isinstance(time, pd.DatetimeIndex):
                self._time = pd.Series(time, copy=False)
            else:
                self._time = cast(pd.Series, time.reset_index(drop=True))
            self._value = value.reset_index(drop=True)
            self._set_univariate_values_to_series()
            # Set time col name
            if time.name:
                self.time_col_name = time.name
            else:
                # pyre-ignore[6]: Expected `Union[typing.Callable[[Optional[typing.Has...
                self._time.rename(DEFAULT_TIME_NAME, inplace=True)
            # Make sure the value series has a name
            if isinstance(self._value, pd.core.series.Series) and self._value.name is None:
                # pyre-ignore[6]: Expected `Union[typing.Callable[[Optional[typing.Has...
                self._value.rename(DEFAULT_VALUE_NAME, inplace=True)
            # Checking for emptiness
            if self.time.empty and self.value.empty:
                logging.warning("Initializing empty TimeSeriesData object")
                self.time = pd.Series([], name=time_col_name)
                if isinstance(value, pd.DataFrame):
                    self.value = pd.Series([], name=DEFAULT_VALUE_NAME)
                else:
                    self.value = pd.Series([], name=value.name if value.name else DEFAULT_VALUE_NAME)
            # Raise exception if only one of time and value is empty
            elif self.time.empty or self.value.empty:
                msg = "One of time or value is empty while the other is not"
                raise _log_error(msg)
            # If time values are passed then standardizing format
            else:
                self.time = cast(
                    pd.Series,
                    self._set_time_format(
                        self.time,
                        date_format=date_format,
                        use_unix_time=use_unix_time,
                        unix_time_units=unix_time_units,
                        tz=tz,
                        tz_ambiguous=tz_ambiguous,
                        tz_nonexistent=tz_nonexistent,
                    ).reset_index(drop=True),
                )

            # Validate that time & value have equal lengths
            self.validate_data(validate_frequency=False, validate_dimension=True)

            df = self.to_dataframe()
            df = self._sort_by_time(sort_by_time=sort_by_time, df=df)
            self._extract_from_df(df=df)

        # If None is passed
        elif not time and not value:
            self._time = pd.Series([], name=time_col_name)
            self._value = pd.Series([], name=DEFAULT_VALUE_NAME)
            logging.warning("Initializing empty TimeSeriesData object")

        # Error if only one of time or value is None
        else:
            msg = "One of time or value is empty while the other is not"
            raise _log_error(msg)

        # Validate values
        if not self.value.empty and not (
            (isinstance(self.value, pd.core.series.Series) and is_numeric_dtype(self.value))
            or (isinstance(self.value, pd.DataFrame) and all(is_numeric_dtype(self.value[col]) for col in self.value))
        ):
            if isinstance(self.value, pd.core.series.Series):
                value_dtypes = self.value.dtypes
            else:  # DataFrame
                value_dtypes = dict(self.value.dtypes)
            msg = f"Time series data is type {value_dtypes} but must be numeric"
            raise _log_error(msg)

        self._calc_min_max_values()

    @property
    def time(self) -> pd.Series:
        """Returns the time values of the series.

        Returns:
          A `pandas.Series` representing the time values of the time series.
        """
        return self._time

    @time.setter
    def time(self, time_values: pd.Series) -> None:
        """Sets the time values of the :class:`TimeSeriesData`.

        Args:
          time_values. A `pandas.Series` with the updated time values.
        """
        self._time = time_values

    @property
    def value(self) -> Union[pd.Series, pd.DataFrame]:
        """Returns the value(s) of the series.

        Returns:
          A `pandas.Series` or `pandas.DataFrame` representing the value(s) of the
          time series.
        """
        return self._value

    @value.setter
    def value(self, values: Union[pd.Series, pd.DataFrame]) -> None:
        """Sets the value(s) of the :class:`TimeSeriesData.`

        Args:
          values: A `pandas.Series` or `pandas.DataFrame` with the updated
          values(s).
        """

        self._value = values
        # updates for min/max values are necessary once values are updated
        self._calc_min_max_values()

    @property
    def min(self) -> Union[pd.Series, float]:
        """Returns the min value(s) of the series.

        Returns:
          A `pandas.Series` or float representing the min value(s) of the
          time series.
        """
        return self._min

    @property
    def max(self) -> Union[pd.Series, float]:
        """Returns the max value(s) of the series.

        Returns:
          A `pandas.Series` or float representing the max value(s) of the
          time series.
        """
        return self._max

    def __eq__(self, other: object) -> bool:
        # Currently "__eq__" only works with other TimeSeriesData objects.
        if not isinstance(other, TimeSeriesData):
            return NotImplemented

        # Check if time values are equal
        try:
            assert_series_equal(self.time, other.time, check_dtype=False)
        except AssertionError:
            return False

        # If both objects are univariate
        if isinstance(self.value, pd.Series) and isinstance(other.value, pd.Series):
            # Check if value Series are equal
            try:
                assert_series_equal(self.value, other.value, check_dtype=False)
            except AssertionError:
                return False
        # If both objects are multivariate
        elif isinstance(self.value, pd.DataFrame) and isinstance(other.value, pd.DataFrame):
            # Check if value DataFrames are equal (ignore column order)
            try:
                assert_frame_equal(
                    self.value.sort_index(axis=1),
                    cast(pd.DataFrame, other.value).sort_index(axis=1),
                    check_names=True,
                    check_dtype=False,
                )
            except AssertionError:
                return False
        # Otherwise one TimeSeriesData is univariate and the other is multivariate
        else:
            return False

        return True

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __sub__(self, other: object) -> TimeSeriesData:
        return self._perform_op(other, OperationsEnum.SUB)

    def __truediv__(self, other: object) -> TimeSeriesData:
        return self._perform_op(other, OperationsEnum.DIV)

    def __add__(self, other: object) -> TimeSeriesData:
        return self._perform_op(other, OperationsEnum.ADD)

    def __mul__(self, other: object) -> TimeSeriesData:
        return self._perform_op(other, OperationsEnum.MUL)

    def __len__(self) -> int:
        return len(self.value)

    def __getitem__(self, sliced: Union[str, Iterable, builtins.slice]) -> TimeSeriesData:
        if isinstance(sliced, str) or (isinstance(sliced, Iterable) and all(isinstance(s, str) for s in sliced)):
            return TimeSeriesData(
                time=self.time,
                value=self.value[sliced],
                time_col_name=self.time_col_name,
            )
        return TimeSeriesData(
            time=self.time[sliced],
            value=self.value[sliced],
            time_col_name=self.time_col_name,
        )

    def __repr__(self) -> str:
        return self.to_dataframe().__repr__()

    def _repr_html_(self) -> str:
        return self.to_dataframe()._repr_html_()

    def _set_univariate_values_to_series(self) -> None:
        # This hack is required since downstream models are expecting value of
        # type Series in case of univariate time series
        if isinstance(self.value, pd.DataFrame) and self.value.shape[1] == 1:
            self.value = self.value.iloc[:, 0]

    def is_empty(self) -> bool:
        """Checks if the :class:`TimeSeriesData` is empty.

        Returns:
          False if :class:`TimeSeriesData` does not have any datapoints.
          Otherwise return True.
        """

        return self.value.empty and self.time.empty

    def _set_time_format(
        self,
        series: pd.Series,
        date_format: Optional[str],
        use_unix_time: Optional[bool],
        unix_time_units: Optional[str],
        tz: Optional[str] = None,
        tz_ambiguous: Union[str, np.ndarray] = "raise",
        tz_nonexistent: str = "raise",
    ) -> pd.core.series.Series:
        """Parses time format when initializing :class:`TimeSeriesData`."""

        # Checking if time column is of type pandas datetime
        if not is_datetime(series):
            if use_unix_time:
                try:
                    if tz:
                        return (
                            pd.to_datetime(series.values, unit=unix_time_units, utc=True)
                            .tz_convert(tz)
                            .to_series()
                            .reset_index(drop=True)
                        )
                    else:
                        return pd.to_datetime(series, unit=unix_time_units)
                except ValueError:
                    msg = "Failed to parse time column " f"{list(series)} using unix units " f"{unix_time_units}"
                    logging.error(msg)
                    raise ValueError(msg)
            # Otherwise try to parse string
            else:
                try:
                    if tz:
                        return (
                            pd.to_datetime(series.values, format=date_format)
                            .tz_localize(tz, ambiguous=tz_ambiguous, nonexistent=tz_nonexistent)
                            .to_series()
                            .reset_index(drop=True)
                        )
                    else:
                        return pd.to_datetime(series, format=date_format)
                except ValueError:
                    msg = "Failed to parse time column " f"{list(series)} using specified format " f"{date_format}"
                    logging.error(msg)
                    raise ValueError(msg)
        else:
            return series

    def extend(self, other: object, validate: bool = True) -> None:
        """
        Extends :class:`TimeSeriesData` with another :class:`TimeSeriesData`
        object.

        Args:
          other: The other :class:`TimeSeriesData` object (currently
            only other :class:`TimeSeriesData` objects are supported).
          validate (optional): A boolean representing if the new
            :class:`TimeSeriesData` should be validated (default True).

        Raises:
          ValueError: The object passed was not an instance of
            :class:`TimeSeriesData`.
        """

        if not isinstance(other, TimeSeriesData):
            raise TypeError("extend must take another TimeSeriesData object")
        # Concatenate times
        self.time = pd.concat([self.time, other.time], ignore_index=True, copy=False).reset_index(drop=True)
        # Convert values to DataFrame if needed
        cur_value = self.value
        other_value = other.value
        if isinstance(self.value, pd.Series):
            cur_value = pd.DataFrame(cur_value, copy=False)
        if isinstance(other.value, pd.Series):
            other_value = pd.DataFrame(other_value, copy=False)
        # Concatenate values
        self.value = pd.concat([cur_value, other_value], ignore_index=True, copy=False).reset_index(drop=True)
        # Merge value back to Series if required
        self._set_univariate_values_to_series()
        # Validate that frequency is constant if required
        if validate:
            self.validate_data(validate_frequency=True, validate_dimension=False)

    def time_to_index(self) -> pd.DatetimeIndex:
        """
        Utility function converting the time in the :class:`TimeSeriesData`
        object to a `pandas.DatetimeIndex`.

        Returns:
          A `pandas.DatetimeIndex` representation of the time values of the series.
        """

        return pd.DatetimeIndex(self.time)

    def validate_data(self, validate_frequency: bool, validate_dimension: bool) -> None:
        """
        Validates the time series for correctness (on both frequency and
        dimension).

        Args:
          validate_frequency: A boolean indicating whether the
            :class:`TimeSeriesData` should be validated for constant frequency.
          validate_dimension: A boolean indicating whether the
            :class:`TimeSeriesData` should be validated for having both the
            same number of timesteps and values.

        Raises:
          ValueError: The frequency and/or dimensions were invalid.
        """

        # check the time frequency is constant
        if validate_frequency and pd.infer_freq(self.time_to_index()) is None:
            raise ValueError("Only constant frequency is supported for time!")

        if validate_dimension and len(self.time) != self.value.shape[0]:
            raise ValueError(
                "time and value have different lengths (dimensions)! " f"({len(self.time)} vs. {self.value.shape[0]})"
            )

    def _calc_min_max_values(self) -> None:
        # Get maximum and minimum values
        if not self.value.empty:
            if isinstance(self.value, pd.core.series.Series):
                self._min = np.nanmin(self.value.values)
                self._max = np.nanmax(self.value.values)
            else:
                self._min = self.value.min(skipna=True)
                self._max = self.value.max(skipna=True)
        else:
            self._min = np.nan
            self._max = np.nan

    def is_data_missing(self) -> bool:
        """
        Checks if data is missing from the time series.

        This is very similar to :meth:`validate_data()` but will not raise an
        error.

        Returns:
          True when data is missing from the time series. Otherwise False.
        """

        # pd.infer_freq needs at least 3 time points.
        # here we tackle the case less than 3 time points
        if len(self.time) < 3:
            return False

        if pd.infer_freq(self.time_to_index()) is None:
            return True
        else:
            return False

    def freq_to_timedelta(self) -> pd.Timedelta:
        """
        Returns a `pandas.Timedelta` representation of the
        :class:`TimeSeriesdata` frequency.

        Returns:
          A `pandas.Timedelta` object representing the frequency of the
          :class:`TimeSeriesData`.
        """

        return pd.Timedelta(to_offset(pd.infer_freq(self.time_to_index())))

    def tz(
        self,
    ) -> Union[datetime.tzinfo, dateutil.tz.tz.tzfile, None]:
        """
        Returns the timezone of the :class:`TimeSeriesData`.

        Returns:
          A timezone aware object representing the timezone of the
          :class:`TimeSeriesData`. Returns None when there is no timezone
          present.

        For more info, see:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.tz.html.
        """

        return self.time_to_index().tz

    def is_univariate(self) -> bool:
        """Returns whether the :class:`TimeSeriesData` is univariate.

        Returns:
          True if the :class:`TimeSeriesData` is univariate. False otherwise.
        """

        return len(self.value.shape) == 1

    def to_dataframe(self, standard_time_col_name: bool = False) -> pd.DataFrame:
        """
        Converts the :class:`TimeSeriesData` object into a `pandas.DataFrame`.

        Args:
          standard_time_col (optional): True if the DataFrame's time column name
            should be "time". To keep the same time column name as the current
            :class:`TimeSeriesData` object, leave as False (default False).
        """

        time_col_name = DEFAULT_TIME_NAME if standard_time_col_name else self.time_col_name
        output_df = pd.DataFrame(dict(zip((time_col_name,), (self.time,))), copy=False)
        if isinstance(self.value, pd.Series):
            if self.value.name is not None:
                output_df[self.value.name] = self.value
            else:
                output_df[DEFAULT_VALUE_NAME] = self.value
        elif isinstance(self.value, pd.DataFrame):
            output_df = pd.concat([output_df, self.value], axis=1, copy=False).reset_index(drop=True)
        else:
            raise ValueError(f"Wrong value type: {type(self.value)}")
        return output_df

    def to_array(self) -> np.ndarray:
        """Converts the :class:`TimeSeriesData` object to a `numpy.ndarray`.

        Returns:
          A `numpy.ndarray` representation of the time series.
        """

        return self.to_dataframe().to_numpy()

    def _get_binary_op_other_arg(self, other: object) -> TimeSeriesData:
        if isinstance(other, float) or isinstance(other, int):
            if isinstance(self.value, pd.Series):
                return TimeSeriesData(
                    pd.DataFrame(
                        dict(
                            zip(
                                (DEFAULT_TIME_NAME, self.value.name),
                                (
                                    self.time,
                                    pd.Series(other, index=self.time.index, copy=False),
                                ),
                            )
                        ),
                        copy=False,
                    )
                )
            else:
                # TODO: implement multivariate time series operation with constant
                raise NotImplementedError("Operation on multivariate")

        if not isinstance(other, TimeSeriesData):
            raise TypeError("Binary op must take another TimeSeriesData object")
        if not self.time.equals(other.time):
            raise ValueError("BBinary op must take a TimeSeriesData with same time")
        return other

    def _perform_op(self, other: object, op_type: "OperationsEnum") -> TimeSeriesData:
        # Extract DataFrames with same time column name for joining
        self_df = self.to_dataframe(standard_time_col_name=True)
        other_df = self._get_binary_op_other_arg(other).to_dataframe(standard_time_col_name=True)
        # Join DataFrames on time column
        combo_df = pd.merge(
            self_df,
            other_df,
            on=DEFAULT_TIME_NAME,
            how="outer",
            suffixes=(PREFIX_OP_1, PREFIX_OP_2),
            copy=False,
        )
        # Map the final column name to the sub column names
        col_map = {}
        for col_name in list(combo_df.columns):
            if PREFIX_OP_1 in col_name:
                prefix = col_name.split(PREFIX_OP_1)[0]
                col_map[prefix] = col_map.get(prefix, []) + [col_name]
            elif PREFIX_OP_2 in col_name:
                prefix = col_name.split(PREFIX_OP_2)[0]
                col_map[prefix] = col_map.get(prefix, []) + [col_name]

        for col_name in list(col_map.keys()):
            # Perform operation on two columns and merge back to one column
            col_1, col_2 = col_map[col_name]
            if op_type == OperationsEnum.ADD:
                combo_df[col_name] = combo_df[col_1] + combo_df[col_2]
            elif op_type == OperationsEnum.SUB:
                combo_df[col_name] = combo_df[col_1] - combo_df[col_2]
            elif op_type == OperationsEnum.DIV:
                combo_df[col_name] = combo_df[col_1] / combo_df[col_2]
            elif op_type == OperationsEnum.MUL:
                combo_df[col_name] = combo_df[col_1] * combo_df[col_2]
            else:
                raise ValueError("Unsupported Operations Type")
            combo_df.drop([col_1, col_2], axis=1, inplace=True)
        # Set columns only present in one of the objects to None
        final_col_list = set([DEFAULT_TIME_NAME] + list(col_map.keys()))
        for col_name in list(combo_df.columns):
            if col_name not in final_col_list:
                combo_df[col_name] = np.nan
        # Change time col name back if needed
        if self.time_col_name != DEFAULT_TIME_NAME:
            combo_df[self.time_col_name] = combo_df[DEFAULT_TIME_NAME]
            combo_df.drop(DEFAULT_TIME_NAME, axis=1, inplace=True)
        return TimeSeriesData(df=combo_df, time_col_name=self.time_col_name)

    def _sort_by_time(self, sort_by_time: bool, df: pd.DataFrame) -> pd.DataFrame:
        """Sort DataFrame by time if necessary, otherwise output warning"""
        if sort_by_time:
            df.sort_values(self.time_col_name, inplace=True)
            df.reset_index(inplace=True, drop=True)
        else:
            logging.warning(("Please make sure the time series is sorted by time or " "set 'sort_by_time' as True."))
        return df

    def _extract_from_df(self, df: pd.DataFrame) -> None:
        """Set instance `time` and `value` variables from DataFrame"""
        self._time = df[self.time_col_name]
        self._value = df[[x for x in df.columns if x != self.time_col_name]]
        self._set_univariate_values_to_series()

    def infer_freq_robust(self) -> pd.Timedelta:
        """
        This method is a more robust way to infer the frequency of the time
        series in the presence of missing data. It looks at the diff of the
        time series, and decides the frequency by majority voting.

        Returns:
          A `pandas.Timedelta` object representing the frequency of the series.

        Raises:
          ValueError: The :class:`TimeSeriesData` has less than 2 data points.
        """

        df = self.to_dataframe()

        if df.shape[0] <= 1:
            raise ValueError("Cannot find frequency for less than two data points")

        freq_counts = df[self.time_col_name].diff().value_counts().sort_values(ascending=False)

        frequency = freq_counts.index[0]

        return frequency

    def interpolate(
        self,
        freq: Optional[Union[str, pd.Timedelta]] = None,
        base: int = 0,
        method: str = "linear",
        remove_duplicate_time: bool = False,
        **kwargs: Any,
    ) -> TimeSeriesData:
        """
        Interpolate missing date if `time` doesn't have constant frequency.

        The following options are available:
          - linear
          - backward fill
          - forward fill
          - all other methods that pd.interpolate supports

        See https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.interpolate.html
        for more detail on these options.

        Args:
          freq: A string representing the pre-defined freq of the time series.
          base: base argument for resample().
            See https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.resample.html
            Note that base will be deprecated since version 1.1.0.
            The new arguments that you should use are ‘offset’ or ‘origin’.
          method: A string representing the method to impute the missing time
            and data. See the above options (default "linear").
          remove_duplicate_index: A boolean to auto-remove any duplicate time
            values, as interpolation in this case due to the need to index
            on time (default False).
          kwargs: additional arguments to pass to pd.interpolate

        Returns:
            A new :class:`TimeSeriesData` object with interpolated data.
        """

        if not freq:
            freq = self.infer_freq_robust()

        # convert to pandas.DataFrame so that we can leverage the built-in methods
        df = self.to_dataframe()

        # Linear interpolation fails if a column has an int type - convert to float
        if method == "linear":
            for col in list(df):
                if col != self.time_col_name:
                    try:
                        df[col] = df[col].astype(float)
                    except ValueError:
                        raise ValueError(f"Column {col} is invalid type: {df[col].dtype}")

        df.set_index(self.time_col_name, inplace=True)

        # Removing duplicate time index values if needed
        if remove_duplicate_time:
            df = df[~df.index.duplicated()]

        if method == "linear":
            df = df.resample(rule=freq, base=base).interpolate(method="linear")

        elif method == "ffill":
            df = df.resample(rule=freq, base=base).ffill()

        elif method == "bfill":
            df = df.resample(rule=freq, base=base).bfill()
        else:
            df = df.resample(rule=freq, base=base).interpolate(method=method, **kwargs)

        df = df.reset_index().rename(columns={"index": self.time_col_name})
        return TimeSeriesData(df, time_col_name=self.time_col_name)

    def plot(
        self,
        cols: Optional[List[str]] = None,
        ax: Optional[plt.Axes] = None,
        grid: bool = True,
        figsize: Optional[FigSize] = None,
        plot_kwargs: Optional[Dict[str, Any]] = None,
        grid_kwargs: Optional[Dict[str, Any]] = None,
    ) -> plt.Axes:
        """Plots the time series.

        Args:
            cols: List of variable names to plot against time. If None,
                plot all variables in the time series data.
            ax: optional Axes to use. If None, create one.
            grid: if True, draw gridlines.
            figsize: if ax is None, the figsize to create. If None, defaults to
                (10, 6).
            plot_kwargs: optional additional arguments to pass to pandas.plot().
            grid_kwargs: optional additional arguments to pass to Axes.grid().
        Returns:
            The matplotlib Axes.
        """
        if self.is_empty():
            raise ValueError("No data to plot")
        # Make sure columns are valid
        df = self.to_dataframe()
        all_cols = list(df.columns)
        all_cols.remove(self.time_col_name)
        if cols is None:
            cols = all_cols
        elif not set(cols).issubset(all_cols):
            logging.error(f"Columns to plot: {cols} are not all in the timeseries")
            raise ValueError("Invalid columns passed")
        if figsize is None:
            figsize = (10, 6)
        if plot_kwargs is None:
            plot_kwargs = {}
        grid_kwargs_ = {"which": "major", "c": "gray", "ls": "-", "lw": 1, "alpha": 0.2}
        if grid_kwargs is not None:
            grid_kwargs_.update(**grid_kwargs)

        # Plot
        logging.info("Plotting time series")
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = plt.gcf()
        if grid:
            ax.grid(True, **grid_kwargs_)
        fig.tight_layout()
        # pyre-ignore[29]: `pd.core.accessor.CachedAccessor` is not a function.
        df.plot(x=self.time_col_name, y=cols, ax=ax, **plot_kwargs)
        return ax


class TimeSeriesIterator:
    a: Optional[pd.DataFrame] = None

    def __init__(self, ts: TimeSeriesData) -> None:
        self.ts: TimeSeriesData = copy.deepcopy(ts)
        self.ts.value = pd.DataFrame(ts.value, copy=False)
        self.start = 0

    def __iter__(self) -> TimeSeriesIterator:
        self.a = pd.DataFrame(
            list(self.ts.value.iloc[:, 0]),
            index=list(self.ts.time),
            columns=["y"],
            copy=False,
        )
        return self

    def __next__(self) -> pd.DataFrame:
        if self.start < self.ts.value.shape[1]:
            x = pd.DataFrame(
                list(self.ts.value.iloc[:, self.start]),
                index=list(self.ts.time),
                columns=["y"],
                copy=False,
            )
            self.start += 1
            return x
        else:
            raise StopIteration


class TSIterator:
    """Iterates through the values of a single timeseries.

    Produces a timeseries with a single point, in case of an
    univariate time series, or a timeseries with an array indicating
    the values at the given location, for a multivariate time series.

    Attributes:
        ts: The input timeseries.
    """

    def __init__(self, ts: TimeSeriesData) -> None:
        self.ts = ts
        self.curr = 0

    def __iter__(self) -> TSIterator:
        return self

    def __next__(self) -> TimeSeriesData:
        if self.curr < len(self.ts.time):
            if self.ts.is_univariate():
                ret = TimeSeriesData(
                    time=pd.Series(self.ts.time[self.curr], copy=False),
                    value=pd.Series(self.ts.value.iloc[self.curr], name=self.curr, copy=False),
                )
            else:
                ret = TimeSeriesData(
                    time=pd.Series(self.ts.time[self.curr], copy=False),
                    value=self.ts.value.loc[[self.curr]],
                )
            self.curr += 1
            return ret
        else:
            raise StopIteration


class Params:
    def __init__(self) -> None:
        pass

    def validate_params(self) -> None:
        pass


class IntervalAnomaly:
    def __init__(
        self,
        start: pd.Timestamp,
        end: pd.Timestamp,
    ) -> None:
        if start >= end:
            raise ValueError("Start value is supposed to be larger than end value.")
        self.start: pd.Timestamp = start
        self.end: pd.Timestamp = end

    @property
    def second_len(self) -> int:
        return (self.end - self.start) / np.timedelta64(1, "s")


@unique
class ModelEnum(Enum):
    """
    This enum lists the options of models to be set for default search space in
    hyper-parameter tuning.
    """

    ARIMA = auto()
    SARIMA = auto()
    HOLTWINTERS = auto()
    LINEAR = auto()
    QUADRATIC = auto()


@unique
class SearchMethodEnum(Enum):
    """
    This enum lists the options of search algorithms to be used in
    hyper-parameter tuning.
    """

    GRID_SEARCH = auto()
    RANDOM_SEARCH_UNIFORM = auto()
    RANDOM_SEARCH_SOBOL = auto()
    BAYES_OPT = auto()


@unique
class OperationsEnum(Enum):
    """
    This enum lists all the mathematical operations that can be performed on
    :class:`TimeSeriesData` objects.
    """

    ADD = auto()
    SUB = auto()
    DIV = auto()
    MUL = auto()


__all__ = [
    "ModelEnum",
    "OperationsEnum",
    "Params",
    "SearchMethodEnum",
    "TimeSeriesChangePoint",
    "TimeSeriesData",
    "TimeSeriesIterator",
    "TSIterator",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Dynamic Nowcasting is a basic model for short-term forecasting.

This modules contains class NowcastingParams, which is the class parameter
and class NowcastingModel, which is the model.

  Typical usage example:

  nr = DyanamicNowcastingModel(data = data, params = NowcastingParams(step = 10))
  nr.feature_extraction()
  nr.label_extraction()
  nr.fit()
  output = nr.predict()
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from typing import Any
from typing import List

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PowerTransformer

import kats.models.model as m
from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.nowcasting.feature_extraction import LAG
from kats.models.nowcasting.feature_extraction import MA
from kats.models.nowcasting.feature_extraction import MOM
from kats.models.nowcasting.feature_extraction import ROC
from kats.models.nowcasting.model_io import deserialize_from_zippy
from kats.models.nowcasting.model_io import serialize_for_zippy


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def poly(df, n):
    """
    Takes the column x from the dataframe df and takes
    the value from x to the power n
    """
    poly = pd.Series(df.x**n, name="poly_" + str(n))
    df = df.join(poly)
    return df


class NowcastingParams(Params):
    """The class for Nowcasting Parameters.

    Takes parameters for class NowcastingModel.

    Attributes:
        step: An integer indicating how many steps ahead we are forecasting. Default is 1.
    """

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, step: int = 1, **kwargs) -> None:
        super().__init__()
        self.step = step
        logging.debug(f"Initialized QuadraticModel with parameters: step:{step}")

    # pyre-fixme[3]: Return type must be annotated.
    def validate_params(self):
        """Raises: NotImplementedError("Subclasses should implement this!")."""

        logging.warning("Method validate_params() is not implemented.")
        raise NotImplementedError("Subclasses should implement this!")


# pyre-fixme[24]: Generic type `m.Model` expects 1 type parameter.
class NowcastingModel(m.Model):
    """The class for Nowcasting Model.

    This class performs data processing and short term prediction, for time series
    based on machine learning methodology.

    Attributes:
        TimeSeriesData: Time Series Data Source.
        NowcastingParams: parameters for Nowcasting.
    """

    def __init__(
        self,
        data: TimeSeriesData,
        params: NowcastingParams,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        model: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        poly_model: Any = None,
        feature_names: List[str] = [],
        poly_feature_names: List[str] = [],
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        scaler: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        label_scaler: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        y_train_season_obj: Any = None,
    ) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: Optional type has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        # pyre-fixme[4]: Attribute must be annotated.
        self.df = data.to_dataframe()
        # pyre-fixme[4]: Attribute must be annotated.
        self.step = params.step
        self.model = model
        self.feature_names = feature_names
        self.poly_model = poly_model
        # pyre-fixme[4]: Attribute must be annotated.
        self.df_poly = data.to_dataframe()
        self.poly_feature_names = poly_feature_names
        # pyre-fixme[4]: Attribute must be annotated.
        self.df_nowcasting = data.to_dataframe()
        self.scaler = scaler
        self.label_scaler = label_scaler
        self.y_train_season_obj = y_train_season_obj

    def feature_extraction(self) -> None:
        """
        Extarct features for training
        """
        # Add the hour, minute, and x column to the data
        self.df_poly["hour"] = self.df_poly["time"].apply(lambda y: y.hour)
        self.df_poly["minute"] = self.df_poly["time"].apply(lambda y: y.minute)
        self.df_poly["x"] = self.df_poly["hour"] * 60 + self.df_poly["minute"]

        # Empty list to hold the feature names
        poly_feature_names = []

        # Add the poly columns to the df_poly
        for degree in [0, 1, 2, 3, 4, 5]:
            self.df_poly = poly(self.df_poly, degree)
            poly_feature_names.append("poly_" + str(degree))

        # filterout + - inf, nan
        self.df_poly = self.df_poly[~self.df_poly.isin([np.nan, np.inf, -np.inf]).any(1)]

        # Save the poly feature name
        self.poly_feature_names = poly_feature_names
        feature_names = []

        #########################################################################################
        train_index_poly = self.df_poly[~self.df_poly.isin([np.nan, np.inf, -np.inf]).any(1)].index
        X_train_poly, y_train_poly = (
            self.df_poly[self.poly_feature_names].loc[train_index_poly],
            self.df_poly["y"].loc[train_index_poly],
        )

        # Build the Polynomial Regression Model
        lin_reg = LinearRegression()
        lin_reg.fit(X_train_poly, y_train_poly)
        self.poly_model = lin_reg
        y_train_season = lin_reg.predict(X_train_poly)
        self.y_train_season_obj = y_train_season
        # print(self.df_poly)
        # print(X_train_poly)
        # print(self.y_train_season_obj)
        #########################################################################################

        # If we have a 0, switch to basic nowcasting!
        if np.sum(self.df["y"].isin([0.0])) > 0:
            for n in [10, 15, 20, 25, 30]:
                self.df = MOM(self.df, n)
                feature_names.append("MOM_" + str(n))
            for n in [10, 15, 20, 25, 30]:
                self.df = LAG(self.df, n)
                feature_names.append("LAG_" + str(n))
            self.df = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)]  # filterout + - inf, nan
            self.feature_names = feature_names

        else:
            for n in [10, 15, 20, 25, 30]:
                self.df = ROC(self.df, n)
                feature_names.append("ROC_" + str(n))
            for n in [10, 15, 20, 25, 30]:
                self.df = MOM(self.df, n)
                feature_names.append("MOM_" + str(n))
            for n in [1, 2, 3, 4, 5, 20]:
                self.df = LAG(self.df, n)
                feature_names.append("LAG_" + str(n))
            for n in [10, 20, 30, 40]:
                self.df = MA(self.df, n)
                feature_names.append("MA_" + str(n))
            self.df = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)]  # filterout + - inf, nan
            self.feature_names = feature_names

    def label_extraction(self) -> None:
        """Extracts labels from time seires data."""

        # self.df["label"] = LAG(self.data.to_dataframe(), -self.step)[
        #    "LAG_-" + str(self.step)
        # ]
        self.df["label"] = self.df["y"]

    ###################### module 1: for offline training ######################

    def fit(self) -> None:
        """Fits model."""

        logging.debug("Call fit() with parameters: " "step:{step}".format(step=self.step))

        if np.sum(self.df["y"].isin([0.0])) == 0:
            n = 1
            train_index = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)].index

            X_train = self.df[self.feature_names].loc[train_index]
            min_max_scaler = preprocessing.MinMaxScaler()
            X_train = min_max_scaler.fit_transform(X_train)
            self.scaler = min_max_scaler

            n = 1
            y_train = (self.df["label"].loc[train_index] - self.y_train_season_obj[train_index]).diff(-n)[:-n]
            pt = PowerTransformer()
            y_train = pt.fit_transform(np.array(y_train).reshape(-1, 1))
            self.label_scaler = pt

            X_train = X_train[:-n]

            # reg = GradientBoostingRegressor()
            reg = RandomForestRegressor()
            # reg =linear_model.Lasso(alpha=10)
            reg.fit(X_train, y_train)
            self.model = reg

        else:
            n = 1
            train_index = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)].index

            X_train = self.df[self.feature_names].loc[train_index]
            # min_max_scaler = preprocessing.MinMaxScaler()
            # X_train = min_max_scaler.fit_transform(X_train)
            # self.scaler = min_max_scaler

            n = 1
            y_train = (self.df["label"].loc[train_index] - self.y_train_season_obj[train_index]).diff(-n)[:-n]
            X_train = X_train[:-n]

            reg = LinearRegression()
            reg.fit(X_train, y_train)
            self.model = reg

    def save_model(self) -> bytes:
        """Saves sklearn model as bytes."""

        return serialize_for_zippy(self.model)

    ###################### module 2: for online prediction ######################
    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def predict(self, **kwargs):
        """Predicts the time series in the future.

        Nowcasting forecasts at the time unit of step ahead.
        This is in order to keep precision and different from usual algorithms.
        If model or df are overwritten in the function, it won't use the internal ones.

        Args:
            model: An external sklearn model.
            df: An external dataset.

        Returns:
            A float variable, the forecast at future step.
        """

        logging.debug("Call predict() with parameters. " "Forecast 1 step only, kwargs:{kwargs}".format(kwargs=kwargs))
        if np.sum(self.df["y"].isin([0.0])) == 0:
            X_test = self.df[-self.step :][self.feature_names]
            X_test = self.scaler.transform(X_test)
            y_predict = self.model.predict(X_test)
            y = self.label_scaler.inverse_transform(y_predict.reshape(-1, 1))
            poly_now = self.y_train_season_obj[-1]
            first_occ = np.where(self.y_train_season_obj == poly_now)
            polynext = self.y_train_season_obj[first_occ[0][0] + self.step]
            now = self.df["y"][-self.step :]
            return (now - poly_now) - y.ravel() + polynext
        else:
            X_test = self.df[-self.step :][self.feature_names]
            y_predict = self.model.predict(X_test)
            poly_now = self.y_train_season_obj[-1]
            first_occ = np.where(self.y_train_season_obj == poly_now)
            polynext = self.y_train_season_obj[first_occ[0][0] + self.step]
            now = self.df["y"][-self.step :]
            # print(y_predict)
            return (now - poly_now) - y_predict + polynext

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def predict_polyfit(self, model=None, df=None, **kwargs):
        poly_now = self.y_train_season_obj[-1]
        first_occ = np.where(self.y_train_season_obj == poly_now)
        polynext = self.y_train_season_obj[first_occ[0][0] + self.step]
        return polynext

    def load_model(self, model_as_bytes: bytes) -> None:
        """Loads model_as_str and decodes into the class NowcastingModel.

        Args:
            model_as_bytes: a binary variable, indicating whether to read as bytes.
        """

        self.model = deserialize_from_zippy(model_as_bytes)

    # pyre-fixme[14]: `plot` overrides method defined in `Model` inconsistently.
    # pyre-fixme[3]: Return type must be annotated.
    def plot(self):
        """Raises: NotImplementedError("Subclasses should implement this!")。"""

        raise NotImplementedError("Subclasses should implement this!")

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        """Returns the name as Nowcasting,"""

        return "Nowcasting"
